import datetime
import time
from collections import deque
from typing import Tuple

from bxcommon import constants
from bxutils import logging
from bxcommon.messages.bloxroute import compact_block_short_ids_serializer
from bxcommon.services.transaction_service import TransactionService
from bxgateway.abstract_message_converter import finalize_block_bytes
from bxcommon.utils import convert, crypto
from bxcommon.utils.object_hash import Sha256Hash
from bxgateway.abstract_message_converter import BlockDecompressionResult
from bxgateway.messages.eth.eth_abstract_message_converter import EthAbstractMessageConverter, parse_block_message
from bxgateway.messages.eth.internal_eth_block_info import InternalEthBlockInfo
from bxgateway.utils.block_info import BlockInfo
from bxcommon.utils.blockchain_utils.eth import rlp_utils, eth_common_utils

logger = logging.get_logger(__name__)


class EthNormalMessageConverter(EthAbstractMessageConverter):

    def block_to_bx_block(
        self,
        block_msg: InternalEthBlockInfo,
        tx_service: TransactionService,
        enable_block_compression: bool,
        min_tx_age_seconds: float
    ) -> Tuple[memoryview, BlockInfo]:
        """
        Convert Ethereum new block message to internal broadcast message with transactions replaced with short ids

        The code is optimized and does not make copies of bytes

        :param block_msg: Ethereum new block message
        :param tx_service: Transactions service
        :param enable_block_compression
        :param min_tx_age_seconds
        :return: Internal broadcast message bytes (bytearray), tuple (txs count, previous block hash)
        """

        compress_start_datetime = datetime.datetime.utcnow()
        compress_start_timestamp = time.time()

        txs_bytes, block_hdr_full_bytes, remaining_bytes, prev_block_bytes = parse_block_message(block_msg)

        used_short_ids = []

        # creating transactions content
        content_size = 0
        buf = deque()
        ignored_sids = []

        tx_start_index = 0
        tx_count = 0
        original_size = len(block_msg.rawbytes())
        max_timestamp_for_compression = time.time() - min_tx_age_seconds

        while True:
            if tx_start_index >= len(txs_bytes):
                break

            _, tx_item_length, tx_item_start = rlp_utils.consume_length_prefix(txs_bytes, tx_start_index)
            tx_bytes = txs_bytes[tx_start_index:tx_item_start + tx_item_length]
            tx_hash_bytes = eth_common_utils.keccak_hash(tx_bytes)
            tx_hash = Sha256Hash(tx_hash_bytes)
            tx_key = tx_service.get_transaction_key(tx_hash)
            short_id = tx_service.get_short_id_by_key(tx_key)
            short_id_assign_time = 0

            if short_id != constants.NULL_TX_SID:
                short_id_assign_time = tx_service.get_short_id_assign_time(short_id)

            if short_id <= constants.NULL_TX_SID or \
                    not enable_block_compression or short_id_assign_time > max_timestamp_for_compression:
                if short_id > constants.NULL_TX_SID:
                    ignored_sids.append(short_id)
                is_full_tx_bytes = rlp_utils.encode_int(1)
                tx_content_bytes = tx_bytes
            else:
                is_full_tx_bytes = rlp_utils.encode_int(0)
                used_short_ids.append(short_id)
                tx_content_bytes = bytes()

            tx_content_prefix = rlp_utils.get_length_prefix_str(len(tx_content_bytes))

            short_tx_content_size = len(is_full_tx_bytes) + len(tx_content_prefix) + len(tx_content_bytes)

            short_tx_content_prefix_bytes = rlp_utils.get_length_prefix_list(short_tx_content_size)

            buf.append(short_tx_content_prefix_bytes)
            buf.append(is_full_tx_bytes)
            buf.append(tx_content_prefix)
            buf.append(tx_content_bytes)

            content_size += len(short_tx_content_prefix_bytes) + short_tx_content_size

            tx_start_index = tx_item_start + tx_item_length

            tx_count += 1

        list_of_txs_prefix_bytes = rlp_utils.get_length_prefix_list(content_size)
        buf.appendleft(list_of_txs_prefix_bytes)
        content_size += len(list_of_txs_prefix_bytes)

        buf.appendleft(block_hdr_full_bytes)
        content_size += len(block_hdr_full_bytes)

        buf.append(remaining_bytes)
        content_size += len(remaining_bytes)

        compact_block_msg_prefix = rlp_utils.get_length_prefix_list(content_size)
        buf.appendleft(compact_block_msg_prefix)
        content_size += len(compact_block_msg_prefix)

        block = finalize_block_bytes(buf, content_size, used_short_ids)
        bx_block_hash = convert.bytes_to_hex(crypto.double_sha256(block))

        block_info = BlockInfo(
            block_msg.block_hash(),
            used_short_ids,
            compress_start_datetime,
            datetime.datetime.utcnow(),
            (time.time() - compress_start_timestamp) * 1000,
            tx_count,
            bx_block_hash,
            convert.bytes_to_hex(prev_block_bytes),
            original_size,
            content_size,
            100 - float(content_size) / original_size * 100,
            ignored_sids
        )

        return memoryview(block), block_info
    
    def bx_block_to_block(self, bx_block_msg, tx_service) -> BlockDecompressionResult:
        """
        Converts internal broadcast message to Ethereum new block message

        The code is optimized and does not make copies of bytes

        :param bx_block_msg: internal broadcast message bytes
        :param tx_service: Transactions service
        :return: tuple (new block message, block hash, unknown transaction short id, unknown transaction hashes)
        """

        if not isinstance(bx_block_msg, (bytearray, memoryview)):
            raise TypeError("Type bytearray is expected for arg block_bytes but was {0}"
                            .format(type(bx_block_msg)))

        decompress_start_datetime = datetime.datetime.utcnow()
        decompress_start_timestamp = time.time()

        block_msg_bytes = bx_block_msg if isinstance(bx_block_msg, memoryview) else memoryview(bx_block_msg)

        block_offsets = compact_block_short_ids_serializer.get_bx_block_offsets(bx_block_msg)
        short_ids, short_ids_bytes_len = compact_block_short_ids_serializer.deserialize_short_ids_from_buffer(
            bx_block_msg,
            block_offsets.short_id_offset
        )

        block_bytes = block_msg_bytes[block_offsets.block_begin_offset: block_offsets.short_id_offset]

        _, block_itm_len, block_itm_start = rlp_utils.consume_length_prefix(block_bytes, 0)
        block_itm_bytes = block_bytes[block_itm_start:]

        _, block_hdr_len, block_hdr_start = rlp_utils.consume_length_prefix(block_itm_bytes, 0)
        full_hdr_bytes = block_itm_bytes[0:block_hdr_start + block_hdr_len]

        block_hash_bytes = eth_common_utils.keccak_hash(full_hdr_bytes)
        block_hash = Sha256Hash(block_hash_bytes)


        _, block_txs_len, block_txs_start = rlp_utils.consume_length_prefix(
            block_itm_bytes, block_hdr_start + block_hdr_len
        )
        txs_bytes = block_itm_bytes[block_txs_start:block_txs_start + block_txs_len]

        remaining_bytes = block_itm_bytes[block_txs_start + block_txs_len:]

        # parse statistics variables
        short_tx_index = 0
        unknown_tx_sids = []
        unknown_tx_hashes = []

        # creating transactions content
        content_size = 0
        buf = deque()
        tx_count = 0

        tx_start_index = 0

        while True:
            if tx_start_index >= len(txs_bytes):
                break

            _, tx_itm_len, tx_itm_start = rlp_utils.consume_length_prefix(txs_bytes, tx_start_index)
            tx_bytes = txs_bytes[tx_itm_start:tx_itm_start + tx_itm_len]

            is_full_tx_start = 0
            is_full_tx, is_full_tx_len, = rlp_utils.decode_int(tx_bytes, is_full_tx_start)

            _, tx_content_len, tx_content_start = rlp_utils.consume_length_prefix(
                tx_bytes, is_full_tx_start + is_full_tx_len)
            tx_content_bytes = tx_bytes[tx_content_start:tx_content_start + tx_content_len]
            if is_full_tx:
                tx_bytes = tx_content_bytes
            else:
                short_id = short_ids[short_tx_index]
                tx_hash, tx_bytes, _ = tx_service.get_transaction(short_id)

                if tx_hash is None:
                    unknown_tx_sids.append(short_id)
                elif tx_bytes is None:
                    unknown_tx_hashes.append(tx_hash)

                short_tx_index += 1

            if tx_bytes is not None and not unknown_tx_sids and not unknown_tx_hashes:
                buf.append(tx_bytes)
                content_size += len(tx_bytes)

            tx_count += 1

            tx_start_index = tx_itm_start + tx_itm_len

        if not unknown_tx_sids and not unknown_tx_hashes:

            txs_prefix = rlp_utils.get_length_prefix_list(content_size)
            buf.appendleft(txs_prefix)
            content_size += len(txs_prefix)

            buf.appendleft(full_hdr_bytes)
            content_size += len(full_hdr_bytes)

            buf.append(remaining_bytes)
            content_size += len(remaining_bytes)

            msg_len_prefix = rlp_utils.get_length_prefix_list(content_size)
            buf.appendleft(msg_len_prefix)

            block_msg_bytes = bytearray(content_size)
            off = 0
            for blob in buf:
                next_off = off + len(blob)
                block_msg_bytes[off:next_off] = blob
                off = next_off

            block_msg = InternalEthBlockInfo(block_msg_bytes)
            logger.debug("Successfully parsed block broadcast message. {} "
                         "transactions in block {}", tx_count, block_hash)

            bx_block_hash = convert.bytes_to_hex(crypto.double_sha256(bx_block_msg))
            compressed_size = len(bx_block_msg)

            block_info = BlockInfo(
                block_hash,
                short_ids,
                decompress_start_datetime,
                datetime.datetime.utcnow(),
                (time.time() - decompress_start_timestamp) * 1000,
                tx_count,
                bx_block_hash,
                convert.bytes_to_hex(block_msg.prev_block_hash().binary),
                len(block_msg.rawbytes()),
                compressed_size,
                100 - float(compressed_size) / content_size * 100,
                []
            )

            return BlockDecompressionResult(block_msg, block_info, unknown_tx_sids, unknown_tx_hashes)
        else:
            logger.debug(
                "Block recovery needed for {}. Missing {} sids, {} tx hashes. "
                "Total txs in block: {}",
                block_hash,
                len(unknown_tx_sids),
                len(unknown_tx_hashes),
                tx_count
            )

            return BlockDecompressionResult(
                None,
                BlockInfo(
                    block_hash,
                    short_ids,
                    decompress_start_datetime, datetime.datetime.utcnow(),
                    (time.time() - decompress_start_timestamp) * 1000,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    []
                ),
                unknown_tx_sids,
                unknown_tx_hashes
            )



from typing import Union, Tuple, Optional, Type
from Crypto.Hash import keccak
import blxr_rlp as rlp

from bxcommon.messages.bloxroute.tx_message import TxMessage
from bxcommon.messages.eth.serializers.transaction_type import EthTransactionType
from bxcommon.models.transaction_flag import TransactionFlag
from bxcommon.utils.blockchain_utils.eth import eth_common_constants, rlp_utils
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon import constants



def raw_tx_to_bx_tx(
    tx_bytes: Union[bytearray, memoryview],
    tx_start_index: int,
    network_num: int,
    transaction_flag: Optional[TransactionFlag] = None,
    account_id: str = constants.DECODED_EMPTY_ACCOUNT_ID
) -> Tuple[TxMessage, Type, int, int]:
    if isinstance(tx_bytes, bytearray):
        tx_bytes = memoryview(tx_bytes)
    tx_item_type, tx_item_length, tx_item_start = rlp_utils.consume_length_prefix(tx_bytes, tx_start_index)
    tx_bytes = tx_bytes[tx_start_index:tx_item_start + tx_item_length]

    if tx_item_type == str:
        tx_hash_bytes = keccak_hash(rlp.decode(tx_bytes.tobytes()))
    else:
        tx_hash_bytes = keccak_hash(tx_bytes)
    msg_hash = Sha256Hash(tx_hash_bytes)
    bx_tx = TxMessage(
        message_hash=msg_hash,
        network_num=network_num,
        tx_val=tx_bytes,
        transaction_flag=transaction_flag,
        account_id=account_id
    )
    return bx_tx, tx_item_type, tx_item_length, tx_item_start


def tx_type(tx_bytes: memoryview, tx_start_index: int) -> Tuple[EthTransactionType, memoryview]:
    """
    Reads the transaction type off the top of the memoryview, then normalizes the
    memoryview so that the beginning of the transaction is at the beginning of
    the returned slice.
    """
    transaction_type = EthTransactionType.LEGACY
    first_byte = tx_bytes[tx_start_index]
    if first_byte <= eth_common_constants.MAX_TRANSACTION_TYPE:
        tx_start_index += 1
        try:
            transaction_type = EthTransactionType(first_byte)
        except ValueError:
            pass
    return transaction_type, tx_bytes[tx_start_index:]


def raw_tx_gas_price(tx_bytes: memoryview, tx_start_index: int) -> int:
    tx_item_type, tx_item_length, tx_item_start = rlp_utils.consume_length_prefix(tx_bytes, tx_start_index)

    # if transaction is an EIP-2718 envelope, decode the string a second time to
    # get the actual tx item
    if tx_item_type == str:
        tx_bytes = tx_bytes[tx_item_start:]
        tx_start_index = 0
        _, tx_item_length, tx_item_start = rlp_utils.consume_length_prefix(tx_bytes, tx_start_index)

    transaction_type, tx_bytes = tx_type(tx_bytes, tx_start_index)
    _, tx_item_length, tx_item_start = rlp_utils.consume_length_prefix(tx_bytes, 0)

    tx_bytes = tx_bytes[tx_item_start:tx_item_start + tx_item_length]

    offset = 0
    if transaction_type == EthTransactionType.ACCESS_LIST:
        _chain_id, chain_id_length = rlp_utils.decode_int(tx_bytes, offset)
        offset += chain_id_length

        _nonce, nonce_length = rlp_utils.decode_int(tx_bytes, offset)
        offset += nonce_length
    else:
        # gas_price is the second field, need to skip the first field (nonce)
        _nonce, nonce_length = rlp_utils.decode_int(tx_bytes, offset)
        offset += nonce_length

    gas_price, _ = rlp_utils.decode_int(tx_bytes, offset)
    return gas_price


def keccak_hash(body: memoryview) -> bytes:
    """
    Ethereum Crypto Utils:
    Calculates SHA3 hash of the string
    :param body: bytes to calculate hash from
    :return: SHA3 hash
    """

    if not body:
        raise ValueError("Input is required")

    k_hash = keccak.new(digest_bits=eth_common_constants.SHA3_LEN_BITS, data=body)
    return k_hash.digest()


def int_to_big_endian(value: int) -> bytes:
    return value.to_bytes((value.bit_length() + 7) // 8 or 1, byteorder="big")


def block_header_number(full_block_header_bytes: Union[memoryview, bytearray]) -> int:
    _, block_header_len, block_header_start = rlp_utils.consume_length_prefix(full_block_header_bytes, 0)
    block_header_bytes = full_block_header_bytes[block_header_start:block_header_start + block_header_len]
    offset = eth_common_constants.FIXED_LENGTH_FIELD_OFFSET
    _difficulty, difficulty_length = rlp_utils.decode_int(block_header_bytes, offset)
    offset += difficulty_length
    number, _ = rlp_utils.decode_int(block_header_bytes, offset)
    return number


def block_header_difficulty(full_block_header_bytes: Union[memoryview, bytearray]) -> int:
    _, block_header_len, block_header_start = rlp_utils.consume_length_prefix(full_block_header_bytes, 0)
    block_header_bytes = full_block_header_bytes[block_header_start:block_header_start + block_header_len]
    offset = eth_common_constants.FIXED_LENGTH_FIELD_OFFSET
    difficulty, _difficulty_length = rlp_utils.decode_int(block_header_bytes, offset)
    return difficulty



import json
import os
import subprocess
import time
from distutils.core import setup
from distutils.sysconfig import get_python_lib
from typing import Tuple, Optional, List

import distro
import requests
import setuptools
from setuptools.command.install import install

ENABLE_DEBUG = True if "ENABLE_DEBUG" in os.environ and os.environ["ENABLE_DEBUG"].upper() == "TRUE" else False

SUPPORTED_OS = {"centos": ["7"], "ubuntu": ["18.04"], "alpine": ["3.8", "3.11"]}

EXTENSIONS_URL = "https://s3.amazonaws.com/downloads.blxrbdn.com/bxextensions/{}/{}-{}/{}"
EXTENSIONS_FILENAME = "bxextensions_{}_{}-{}.tar.gz"

MANIFEST_FILENAME = "MANIFEST.MF"
SCRIPT_FILENAME = "bloxroute-gateway"

README_FILENAME = "README.md"
REQUIREMENTS_FILENAME = "requirements.txt"

BXGATEWAY_DIRECTORY = os.path.join("bxgateway", "src", "bxgateway")
BXGATEWAY_CLI_DIRECTORY = os.path.join("bxgateway", "src", "bloxroute_cli")
BXCOMMON_DIRECTORY = os.path.join("bxcommon", "src", "bxcommon")
BXUTILS_DIRECTORY = os.path.join("bxcommon", "src", "bxutils")

_log_messages = []


def log(message: str):
    global _log_messages
    timestamp = time.strftime("%Y-%m-%d-%H:%M:%S+0000-", time.gmtime())
    _log_messages.append(f"{timestamp}: {message}")


def get_distro_properties() -> Tuple[Optional[str], Optional[str]]:
    distro_os_flavor, distro_os_version, _ = distro.linux_distribution()
    log(f"Detected operating system: {distro_os_flavor}: {distro_os_version}.")

    os_flavor = distro_os_flavor.lower().split()[0]
    if os_flavor in SUPPORTED_OS:
        supported_versions = [os_version
                              for os_version in SUPPORTED_OS[os_flavor]
                              if os_version in distro_os_version]
        if supported_versions:
            assert len(supported_versions) == 1
            return os_flavor, supported_versions[0]

    log(f"Your operating system ({os_flavor}:{distro_os_version}) does not support using C++ extensions. "
        f"Setting bloxroute-gateway to run without them.")
    return None, None


def get_version(src_dir: str) -> str:
    manifest_path = os.path.join(src_dir, MANIFEST_FILENAME)
    try:
        with open(manifest_path) as manifest_file:
            manifest_data = json.load(manifest_file)
        full_version = manifest_data["source_version"]
        trimmed_version = ".".join(full_version.split(".")[:3])
        return trimmed_version
    except (IOError, json.JSONDecodeError) as e:
        raise SystemError(f"Could not read manifest file at: {manifest_path}. Error: {e}")


def get_requirements(src_dir: str) -> List[str]:
    requirements_path = os.path.join(src_dir, REQUIREMENTS_FILENAME)
    requirements = []
    with open(requirements_path) as requirements_file:
        for requirement_line in requirements_file:
            if requirement_line != "\n":
                requirements.append(requirement_line)
    return requirements


def download_extensions(build_path: str, gateway_version: str, os_flavor: str, os_version: str) -> str:
    file_name = EXTENSIONS_FILENAME.format(gateway_version, os_flavor, os_version)
    url = EXTENSIONS_URL.format(gateway_version, os_flavor, os_version, file_name)
    download_file_path = os.path.abspath(os.path.join(build_path, file_name))
    log(f"Downloading C++ extensions from: {url} to {download_file_path}.")

    try:
        download_response = requests.get(url, allow_redirects=False)
        with open(download_file_path, "wb") as download_file:
            download_file.write(download_response.content)
    except Exception as ex:
        raise SystemError(f"Could not download C++ extensions tar.gz, due to {ex}.")

    return download_file_path


def extract_extensions(extensions_file_path: str, build_path: str):
    """
    Extracts C++ extensions from tar ball. Do not extract files to bxextensions/ folder, since
    that would force imports to be for e.g. `bxextensions.task_pool_executor`.

    .so files will be installed at python/site-packages directly.
    """
    extract_path = os.path.abspath(build_path)

    try:
        os.makedirs(extract_path, exist_ok=True)
        subprocess.check_call(["tar", "xvzf", extensions_file_path], cwd=extract_path)
        log(f"C++ extensions have been extracted to {extract_path}.")
    except Exception as ex:
        raise SystemError(f"Failed to extract C++ extensions, due to {ex}.")


def prepare_entrypoint_script(build_path: str, use_extensions: bool):
    script_file_path = os.path.join(build_path, SCRIPT_FILENAME)

    script_contents = ""
    script_contents += f"#!/usr/bin/env bash {os.linesep}\n"
    if os.getenv("LD_LIBRARY_PATH") is None:
        script_contents += f"export LD_LIBRARY_PATH={get_python_lib()}\n"
        os.environ["LD_LIBRARY_PATH"] = get_python_lib()
    else:
        script_contents += f"export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:{get_python_lib()}\n"
    script_contents += f"python3 -c 'from bxgateway.main import main; main()'" \
        f" --use-extensions {use_extensions} $* {os.linesep}"

    try:
        with open(script_file_path, "w") as file_handler:
            file_handler.write(script_contents)

        file_permissions = os.stat(script_file_path)
        os.chmod(script_file_path, file_permissions.st_mode | 0o111)
    except Exception as ex:
        raise RuntimeError(f"Failed to write entry point script to {script_file_path}, due to {ex}. "
                           f"Contact support@bloxroute.com for assistance.")


class ExtensionsInstall(install):
    def run(self):
        install.run(self)

        version = get_version(os.path.join("bxgateway", "src", "bxgateway"))
        os_flavor, os_version = get_distro_properties()

        use_extensions = os_flavor is not None and os_version is not None
        if use_extensions:
            try:
                extensions_file_path = download_extensions(self.build_lib, version, os_flavor, os_version)
                extract_extensions(extensions_file_path, self.install_lib)
            except SystemError as e:
                log(f"Extensions installation failure: {e}. Setting system to run without extensions.")
                use_extensions = False
            except RuntimeError as e:
                log(f"Extensions import failed: {e}. Contact support@bloxroute.com for assistance. "
                    f"Setting system to run without extensions.")
                use_extensions = False

        prepare_entrypoint_script(self.build_lib, use_extensions)

        # copy script to executable path
        if not os.path.isdir(self.install_scripts):
            os.makedirs(self.install_scripts)

        source = os.path.join(self.build_lib, SCRIPT_FILENAME)
        target = os.path.join(self.install_scripts, SCRIPT_FILENAME)
        if os.path.isfile(target):
            os.remove(target)

        self.copy_file(source, target)

        if ENABLE_DEBUG:
            log_dump = os.path.join(self.install_scripts, f"bloxroute-gateway-install-{version}.log")
            with open(log_dump, "w") as f:
                f.write("\n".join(_log_messages) + "\n")


def main():
    version = get_version(BXGATEWAY_DIRECTORY)

    requirements = get_requirements(os.path.join("bxgateway")) + get_requirements(
        os.path.join("bxcommon"))

    packages = []
    # TODO: exclude bxcommon.test_utils* –– need to remove dependency on mock class in real code
    packages.extend(setuptools.find_packages("bxcommon/src",
                                             exclude=["bxcommon.test_utils"]))
    # TODO: exclude bxgateway.testing* –– need to remove dependency on testing class in real code
    packages.extend(setuptools.find_packages("bxgateway/src",
                                             exclude=["bxgateway.testing.mocks"]))

    with open(os.path.join("bxgateway", "README.gateway.md")) as f:
        readme = f.read()

    setup(
        name="bloxroute-gateway",
        version=version,
        url="https://bloxroute.com",
        author="bloXroute Labs",
        author_email="support@bloxroute.com",
        description="bloXroute network gateway for scaling blockchains",
        license="MIT",
        long_description=readme,
        long_description_content_type="text/markdown",
        packages=packages,
        install_requires=requirements,
        include_package_data=True,
        package_dir={
            "bxcommon": BXCOMMON_DIRECTORY,
            "bxutils": BXUTILS_DIRECTORY,
            "bxgateway": BXGATEWAY_DIRECTORY,
            "bloxroute_cli": BXGATEWAY_CLI_DIRECTORY
        },
        python_requires="~=3.6",
        classifiers=[
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Development Status :: 4 - Beta",
            "License :: OSI Approved :: MIT License",
        ],
        cmdclass={
            "install": ExtensionsInstall,
        },
        entry_points={
            "console_scripts": ["bloxroute-cli=bloxroute_cli.main:run_main"]
        }
    )


if __name__ == "__main__":
    main()



import struct
from enum import Enum
from typing import Tuple, List, Optional, Union

from bxcommon.utils.blockchain_utils.ont.ont_object_hash import OntObjectHash
from bxgateway import ont_constants
from bxgateway.messages.ont.ont_message import OntMessage
from bxgateway.messages.ont.ont_message_type import OntMessageType


class InventoryOntType(Enum):
    MSG_TX = 1
    MSG_BLOCK = 2
    MSG_CONSENSUS = 224


class InvOntMessage(OntMessage):
    MESSAGE_TYPE = OntMessageType.INVENTORY

    def __init__(self, magic: Optional[int] = None, inv_type: Optional[Union[InventoryOntType, bytes]] = None,
                 blocks: Optional[List[OntObjectHash]] = None, buf: Optional[bytearray] = None):
        if buf is None:
            buf = bytearray(ont_constants.ONT_HDR_COMMON_OFF + ont_constants.ONT_CHAR_LEN + ont_constants.ONT_INT_LEN +
                            # pyre-fixme[6]: Expected `Sized` for 1st param but got
                            #  `Optional[List[OntObjectHash]]`.
                            ont_constants.ONT_HASH_LEN * len(blocks))
            self.buf = buf
            # pyre-fixme[6]: Expected `Sized` for 1st param but got
            #  `Optional[List[OntObjectHash]]`.
            self.blocks_len = len(blocks)

            off = ont_constants.ONT_HDR_COMMON_OFF
            if isinstance(inv_type, InventoryOntType):
                struct.pack_into("<B", buf, off, inv_type.value)
            else:
                # pyre-fixme[6]: Expected `Union[typing.Iterable[int], bytes]` for
                #  2nd param but got `Optional[bytes]`.
                buf[off: off + ont_constants.ONT_CHAR_LEN] = inv_type
            off += ont_constants.ONT_CHAR_LEN
            # pyre-fixme[6]: Expected `Sized` for 1st param but got
            #  `Optional[List[OntObjectHash]]`.
            struct.pack_into("<L", buf, off, len(blocks))
            off += ont_constants.ONT_INT_LEN

            # pyre-fixme[16]: `Optional` has no attribute `__iter__`.
            for block in blocks:
                buf[off:off + ont_constants.ONT_HASH_LEN] = block.get_big_endian()
                off += ont_constants.ONT_HASH_LEN

            super(InvOntMessage, self).__init__(magic, self.MESSAGE_TYPE, off - ont_constants.ONT_HDR_COMMON_OFF, buf)

        else:
            self.buf = buf
            self.blocks_len = 0 if blocks is None else len(blocks)
            self._memoryview = memoryview(buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None

        self._inv_type: Optional[int] = None
        self._block_hashes: Optional[List[OntObjectHash]] = None
        self._len_blocks: Optional[int] = None

    def __repr__(self):
        if self._inv_type is None:
            self.inv_type()
        return "InventoryOntMessage<length:{}, inventory type: {}, number of blocks:{}, blocks: {}>".format(
            len(self.rawbytes()), self._inv_type, self._len_blocks, self._block_hashes)

    def inv_type(self) -> Tuple[int, List[OntObjectHash]]:
        if self._inv_type is None:
            off = ont_constants.ONT_HDR_COMMON_OFF
            self._inv_type, = struct.unpack_from("<B", self.buf, off)
            off += ont_constants.ONT_CHAR_LEN
            self._len_blocks, = struct.unpack_from("<L", self.buf, off)
            off += ont_constants.ONT_INT_LEN
            self._block_hashes = []
            len_blocks = self._len_blocks
            assert isinstance(len_blocks, int)
            for _ in range(len_blocks):
                # pyre-fixme[16]: `Optional` has no attribute `append`.
                self._block_hashes.append(OntObjectHash(buf=self.buf, offset=off, length=ont_constants.ONT_HASH_LEN))
                off += ont_constants.ONT_HASH_LEN

        inv_type, block_hashes = self._inv_type, self._block_hashes
        assert isinstance(inv_type, int)
        assert isinstance(block_hashes, list)
        return inv_type, block_hashes



import struct
from typing import Optional, List

from bxgateway import ont_constants
from bxgateway.messages.ont.ont_message import OntMessage
from bxgateway.messages.ont.ont_message_type import OntMessageType
from bxgateway.messages.ont.ont_messages_util import ont_varint_to_int


class HeadersOntMessage(OntMessage):
    MESSAGE_TYPE = OntMessageType.HEADERS

    def __init__(self, magic: Optional[int] = None, headers: Optional[list] = None, buf: Optional[bytearray] = None):
        if buf is None:
            # pyre-fixme[6]: Expected `Sized` for 1st param but got
            #  `Optional[List[typing.Any]]`.
            headers_length = len(headers)
            # pyre-fixme[16]: `Optional` has no attribute `__getitem__`.
            header_length = len(headers[0])
            buf = bytearray(ont_constants.ONT_HDR_COMMON_OFF + headers_length * header_length +
                            ont_constants.ONT_INT_LEN)
            self.buf = buf

            off = ont_constants.ONT_HDR_COMMON_OFF
            struct.pack_into("<L", buf, off, headers_length)
            off += ont_constants.ONT_INT_LEN
            # pyre-fixme[16]: `Optional` has no attribute `__iter__`.
            for header in headers:
                buf[off:off + len(header)] = header
                off += len(header)

            super().__init__(magic, self.MESSAGE_TYPE, off - ont_constants.ONT_HDR_COMMON_OFF, buf)
        else:
            self.buf = buf
            self._memoryview = memoryview(self.buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None

        self._headers = None
        self._parsed = False

    def headers(self) -> List[memoryview]:
        if not self._parsed:
            self._parsed = True
            off = ont_constants.ONT_HDR_COMMON_OFF
            headers_length, = struct.unpack_from("<L", self.buf, off)
            off += ont_constants.ONT_INT_LEN
            self._headers = []
            for _ in range(headers_length):
                header_start_index = off
                off += ont_constants.ONT_INT_LEN + 3 * ont_constants.ONT_HASH_LEN + \
                       ont_constants.ONT_BLOCK_TIME_HEIGHT_CONS_DATA_LEN
                consensus_payload_length, size = ont_varint_to_int(self.buf, off)
                off += consensus_payload_length + size
                off += ont_constants.ONT_BLOCK_NEXT_BOOKKEEPER_LEN
                header = self._memoryview[header_start_index:off]
                # pyre-fixme[16]: Optional type has no attribute `append`.
                self._headers.append(header)
                bookkeepers_length, size = ont_varint_to_int(self.buf, off)
                off += size
                for _ in range(bookkeepers_length):
                    _, size = ont_varint_to_int(self.buf, off)
                    off += size + ont_constants.ONT_BOOKKEEPER_LEN
                sig_data_length, size = ont_varint_to_int(self.buf, off)
                off += size
                for _ in range(sig_data_length):
                    sig_length, size = ont_varint_to_int(self.buf, off)
                    off += size + sig_length
        headers = self._headers
        assert isinstance(headers, List)
        return headers



import struct

from bxcommon import constants
from bxcommon.messages.abstract_internal_message import AbstractInternalMessage
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.bdn_performance_stats_message import BdnPerformanceStatsMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.messages.bloxroute.v18.bdn_performance_stats_message_v18 import \
    BdnPerformanceStatsMessageV18
from bxcommon.messages.versioning.abstract_message_converter import AbstractMessageConverter


class _BdnPerformanceStatsMessageConverterV18(AbstractMessageConverter):
    _MSG_TYPE_TO_OLD_MSG_CLASS_MAPPING = {
        BloxrouteMessageType.BDN_PERFORMANCE_STATS: BdnPerformanceStatsMessageV18
    }

    _MSG_TYPE_TO_NEW_MSG_CLASS_MAPPING = {
        BloxrouteMessageType.BDN_PERFORMANCE_STATS: BdnPerformanceStatsMessage
    }

    _BASE_LENGTH = (
        AbstractBloxrouteMessage.HEADER_LENGTH
    )

    _INTERVAL_TIMES_BREAKPOINT = (
        _BASE_LENGTH +
        (2 * constants.DOUBLE_SIZE_IN_BYTES)
    )
    _INTERVAL_TIMES_LENGTH = _INTERVAL_TIMES_BREAKPOINT - _BASE_LENGTH

    _MEMORY_UTILIZATION_BREAKPOINT = (
        _INTERVAL_TIMES_BREAKPOINT +
        constants.UL_SHORT_SIZE_IN_BYTES
    )
    _MEMORY_UTILIZATION_LENGTH = _MEMORY_UTILIZATION_BREAKPOINT - _INTERVAL_TIMES_BREAKPOINT

    _NUM_BLOCKCHAIN_PEERS_BREAKPOINT = (
        _MEMORY_UTILIZATION_BREAKPOINT +
        constants.UL_SHORT_SIZE_IN_BYTES
    )
    _NUM_BLOCKCHAIN_PEERS_LENGTH = constants.UL_SHORT_SIZE_IN_BYTES

    _FIRST_NODE_STATS_BREAKPOINT = (
        _NUM_BLOCKCHAIN_PEERS_BREAKPOINT +
        (2 * constants.UL_SHORT_SIZE_IN_BYTES) +
        (5 * constants.UL_INT_SIZE_IN_BYTES)
    )
    _NODE_STATS_LENGTH = _FIRST_NODE_STATS_BREAKPOINT - _NUM_BLOCKCHAIN_PEERS_BREAKPOINT

    _MESSAGE_LEN_WITHOUT_NODE_STATS = (
        _BASE_LENGTH +
        _INTERVAL_TIMES_LENGTH + _MEMORY_UTILIZATION_LENGTH + _NUM_BLOCKCHAIN_PEERS_LENGTH +
        constants.CONTROL_FLAGS_LEN
    )

    _OLD_MESSAGE_NODE_STATS_LEN = (
        constants.IP_ADDR_SIZE_IN_BYTES + constants.UL_SHORT_SIZE_IN_BYTES +
        (5 * constants.UL_INT_SIZE_IN_BYTES) +
        (2 * constants.UL_SHORT_SIZE_IN_BYTES)
    )
    _NEW_MESSAGE_NODE_STATS_LEN = (
        _OLD_MESSAGE_NODE_STATS_LEN +
        (2 * constants.UL_INT_SIZE_IN_BYTES)
    )
    _NODE_STATS_LENGTH_DIFFERENCE = _NEW_MESSAGE_NODE_STATS_LEN - _OLD_MESSAGE_NODE_STATS_LEN

    def convert_from_older_version(
        self, msg: AbstractInternalMessage
    ) -> AbstractInternalMessage:
        msg_type = msg.MESSAGE_TYPE

        if msg_type not in self._MSG_TYPE_TO_NEW_MSG_CLASS_MAPPING:
            raise ValueError(
                f"Tried to convert unexpected old message type from v18: {msg_type}"
            )
        if len(msg.rawbytes()) == self._MESSAGE_LEN_WITHOUT_NODE_STATS:
            # per Endpoint Stats is not available
            length_difference = 0
            new_message_len = self._MESSAGE_LEN_WITHOUT_NODE_STATS
            num_blockchain_peers = 0
        else:
            num_blockchain_peers, = struct.unpack_from("<I", msg.rawbytes(), self._MEMORY_UTILIZATION_BREAKPOINT)
            length_difference = self._NODE_STATS_LENGTH_DIFFERENCE * num_blockchain_peers
            new_message_len = self._MESSAGE_LEN_WITHOUT_NODE_STATS + \
                (num_blockchain_peers * self._NEW_MESSAGE_NODE_STATS_LEN)

        new_msg_class = self._MSG_TYPE_TO_NEW_MSG_CLASS_MAPPING[msg_type]
        new_payload_len = msg.payload_len() + length_difference

        new_msg_bytes = bytearray(new_message_len)
        new_msg_bytes[:self._NUM_BLOCKCHAIN_PEERS_BREAKPOINT] = msg.rawbytes()[:self._NUM_BLOCKCHAIN_PEERS_BREAKPOINT]
        new_msg_off = self._NUM_BLOCKCHAIN_PEERS_BREAKPOINT
        old_msg_off = self._NUM_BLOCKCHAIN_PEERS_BREAKPOINT
        default_new_stats = 0
        for _ in range(num_blockchain_peers):
            new_msg_bytes[new_msg_off:new_msg_off + self._OLD_MESSAGE_NODE_STATS_LEN] = \
                msg.rawbytes()[old_msg_off:old_msg_off + self._OLD_MESSAGE_NODE_STATS_LEN]
            new_msg_off += self._OLD_MESSAGE_NODE_STATS_LEN
            old_msg_off += self._OLD_MESSAGE_NODE_STATS_LEN

            struct.pack_into("<I", new_msg_bytes, new_msg_off, default_new_stats)
            new_msg_off += constants.UL_INT_SIZE_IN_BYTES
            struct.pack_into("<I", new_msg_bytes, new_msg_off, default_new_stats)
            new_msg_off += constants.UL_INT_SIZE_IN_BYTES

        new_msg_bytes[new_msg_off:] = msg.rawbytes()[old_msg_off:]

        return AbstractBloxrouteMessage.initialize_class(
            new_msg_class,
            new_msg_bytes,
            (msg_type, new_payload_len)
        )

    def convert_to_older_version(
        self, msg: AbstractInternalMessage
    ) -> AbstractInternalMessage:
        # Message is one-way from gateway to relay, so converting from new to old version is not necessary.
        raise NotImplementedError

    def convert_first_bytes_to_older_version(
        self, first_msg_bytes: memoryview
    ) -> memoryview:
        raise NotImplementedError

    def convert_first_bytes_from_older_version(
        self, first_msg_bytes: memoryview
    ) -> memoryview:
        raise NotImplementedError

    def convert_last_bytes_to_older_version(
        self, last_msg_bytes: memoryview
    ) -> memoryview:
        raise NotImplementedError

    def convert_last_bytes_from_older_version(
        self, last_msg_bytes: memoryview
    ) -> memoryview:
        raise NotImplementedError

    def get_message_size_change_to_older_version(self) -> int:
        # Message is one-way from gateway to relay, so converting from new to old version is not necessary.
        raise NotImplementedError

    def get_message_size_change_from_older_version(self) -> int:
        raise NotImplementedError


bdn_performance_stats_message_converter_v18 = _BdnPerformanceStatsMessageConverterV18()



import time
from abc import abstractmethod
from datetime import datetime
from typing import Optional, List, Tuple, Union

from bxcommon.messages.bloxroute.tx_message import TxMessage
from bxcommon.models.transaction_flag import TransactionFlag
from bxcommon.utils import crypto, convert
from bxcommon.utils.blockchain_utils.bdn_tx_to_bx_tx import bdn_tx_to_bx_tx
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon import constants as common_constants

from bxgateway import ont_constants
from bxgateway.abstract_message_converter import AbstractMessageConverter, BlockDecompressionResult
from bxgateway.messages.ont.block_ont_message import BlockOntMessage
from bxgateway.messages.ont.consensus_ont_message import OntConsensusMessage
from bxgateway.messages.ont.ont_message import OntMessage
from bxgateway.messages.ont.tx_ont_message import TxOntMessage
from bxgateway.utils.block_info import BlockInfo


def get_block_info(
        bx_block: memoryview,
        block_hash: Sha256Hash,
        short_ids: List[int],
        decompress_start_datetime: datetime,
        decompress_start_timestamp: float,
        total_tx_count: Optional[int] = None,
        ont_block_msg: Optional[Union[BlockOntMessage, OntConsensusMessage]] = None
) -> BlockInfo:
    if ont_block_msg is not None:
        bx_block_hash = convert.bytes_to_hex(crypto.double_sha256(bx_block))
        compressed_size = len(bx_block)
        prev_block_hash = convert.bytes_to_hex(ont_block_msg.prev_block_hash().binary)
        ont_block_len = len(ont_block_msg.rawbytes())
        compression_rate = 100 - float(compressed_size) / ont_block_len * 100
    else:
        bx_block_hash = None
        compressed_size = None
        prev_block_hash = None
        ont_block_len = None
        compression_rate = None
    return BlockInfo(
        block_hash,
        short_ids,
        decompress_start_datetime,
        datetime.utcnow(),
        (time.time() - decompress_start_timestamp) * 1000,
        total_tx_count,
        bx_block_hash,
        prev_block_hash,
        ont_block_len,
        compressed_size,
        compression_rate,
        []
    )


class AbstractOntMessageConverter(AbstractMessageConverter):

    def __init__(self, ont_magic: int):
        self._ont_magic = ont_magic

    @abstractmethod
    def block_to_bx_block(
        self, block_msg, tx_service, enable_block_compression: bool, min_tx_age_seconds: float
    ) -> Tuple[memoryview, BlockInfo]:
        """
        Pack a blockchain block's transactions into a bloXroute block.
        """
        pass

    @abstractmethod
    def bx_block_to_block(self, bx_block_msg, tx_service) -> BlockDecompressionResult:
        """
        Uncompresses a bx_block from a broadcast bx_block message and converts to a raw ONT bx_block.

        bx_block must be a memoryview, since memoryview[offset] returns a bytearray, while bytearray[offset] returns
        a byte.
        """
        pass

    # pyre-fixme[14]: `bx_tx_to_tx` overrides method defined in
    #  `AbstractMessageConverter` inconsistently.
    def bx_tx_to_tx(self, tx_msg: TxMessage):
        # pyre-fixme[6]: Expected `bytes` for 1st param but got `memoryview`.
        buf = bytearray(ont_constants.ONT_HDR_COMMON_OFF) + tx_msg.tx_val()
        raw_ont_tx_msg = OntMessage(self._ont_magic, TxOntMessage.MESSAGE_TYPE, len(tx_msg.tx_val()), buf)
        ont_tx_msg = TxOntMessage(buf=raw_ont_tx_msg.buf)

        return ont_tx_msg

    def tx_to_bx_txs(
        self,
        tx_msg,
        network_num: int,
        transaction_flag: Optional[TransactionFlag] = None,
        min_tx_network_fee: int = 0,
        account_id: str = common_constants.DECODED_EMPTY_ACCOUNT_ID
    ) -> List[Tuple[TxMessage, Sha256Hash, Union[bytearray, memoryview]]]:
        bx_tx_msg = TxMessage(
            tx_msg.tx_hash(),
            network_num,
            tx_val=tx_msg.tx(),
            transaction_flag=transaction_flag,
            account_id=account_id
        )
        return [(bx_tx_msg, tx_msg.tx_hash(), tx_msg.tx())]

    def bdn_tx_to_bx_tx(
        self,
        raw_tx: Union[bytes, bytearray, memoryview],
        network_num: int,
        transaction_flag: Optional[TransactionFlag] = None,
        account_id: str = common_constants.DECODED_EMPTY_ACCOUNT_ID
    ) -> TxMessage:
        return bdn_tx_to_bx_tx(raw_tx, network_num, transaction_flag, account_id)



from typing import List, Optional

from bxcommon import constants
from bxcommon.messages.bloxroute import transactions_info_serializer
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.models.transaction_info import TransactionInfo
from bxutils import logging
from bxutils.logging.log_level import LogLevel

logger = logging.get_logger(__name__)


class TxsMessage(AbstractBloxrouteMessage):
    MESSAGE_TYPE = BloxrouteMessageType.TRANSACTIONS
    """
    Message with tx details. Reply to GetTxsMessage.
    """

    # pyre-fixme[9]: buf has type `bytearray`; used as `None`.
    def __init__(self, txs: Optional[List[TransactionInfo]] = None, buf: bytearray = None) -> None:

        """
        Constructor. Expects list of transaction details or message bytes.

        :param txs: tuple with 3 values (tx short id, tx hash, tx contents)
        :param buf: message bytes
        """

        if buf is None:
            assert txs is not None
            buf = self._txs_to_bytes(txs)

        super(TxsMessage, self).__init__(self.MESSAGE_TYPE, len(buf) - self.HEADER_LENGTH, buf)
        self._txs = None

    def log_level(self):
        return LogLevel.DEBUG

    def get_txs(self) -> List[TransactionInfo]:
        if self._txs is None:
            self._parse()

        assert self._txs is not None
        # pyre-fixme[7]: Expected `List[TransactionInfo]` but got `None`.
        return self._txs

    def _txs_to_bytes(self, txs_details: List[TransactionInfo]):
        # msg_size = HDR_COMMON_OFF + tx count + (sid + hash + tx size) of each tx
        msg_size = (
            self.HEADER_LENGTH
            + transactions_info_serializer.get_serialized_length(txs_details)
            + constants.CONTROL_FLAGS_LEN
        )

        buf = bytearray(msg_size)
        off = self.HEADER_LENGTH

        transactions_info_serializer.serialize_transactions_info_to_buffer(txs_details, buf, off)

        return buf

    def _parse(self):
        self._txs, _ = transactions_info_serializer.deserialize_transactions_info_from_buffer(
            self.buf,
            self.HEADER_LENGTH
        )

    def __repr__(self):
        return "TxsMessage<num_txs: {}>".format(len(self.get_txs()))



import blxr_rlp as rlp
from typing import List, Tuple

from bxutils.logging.log_level import LogLevel

from bxcommon.messages.abstract_block_message import AbstractBlockMessage
from bxcommon.utils.object_hash import Sha256Hash

from bxgateway.messages.eth.protocol.eth_protocol_message import EthProtocolMessage
from bxgateway.messages.eth.protocol.eth_protocol_message_type import EthProtocolMessageType
from bxgateway.messages.eth.serializers.block_hash import BlockHash
from bxcommon.utils.blockchain_utils.eth import rlp_utils


class NewBlockHashesEthProtocolMessage(EthProtocolMessage, AbstractBlockMessage):
    msg_type = EthProtocolMessageType.NEW_BLOCK_HASHES

    fields = [("block_hash_number_pairs", rlp.sedes.CountableList(BlockHash))]

    def __repr__(self) -> str:
        return f"NewBlockHashesEthProtocolMessage<block_hashes: [{self.get_block_hash_number_pairs()}]"

    def block_hash(self) -> Sha256Hash:
        """
        Returns the first block hash in the message

        This is used when send new block hash to the blockchain node. Always send a single block in that message.
        :return:
        """

        block_hash_number_pairs = self.get_field_value("block_hash_number_pairs")
        return Sha256Hash(block_hash_number_pairs[0].hash)

    def timestamp(self) -> int:
        raise NotImplementedError()

    def extra_stats_data(self) -> str:
        return "Block hash"

    def get_block_hash_number_pairs(self) -> List[Tuple[Sha256Hash, int]]:
        return [(Sha256Hash(block_hash.hash), block_hash.number) for block_hash in
                self.get_field_value("block_hash_number_pairs")]

    @classmethod
    def from_block_hash_number_pair(cls, block_hash: Sha256Hash, number: int) -> "NewBlockHashesEthProtocolMessage":
        block_hash_bytes = block_hash.binary
        msg_size = len(block_hash_bytes)

        block_hash_prefix = rlp_utils.get_length_prefix_str(len(block_hash_bytes))
        msg_size += len(block_hash_prefix)

        number_bytes = rlp_utils.encode_int(number)
        msg_size += len(number_bytes)

        block_hash_and_number_prefix = rlp_utils.get_length_prefix_list(msg_size)
        msg_size += len(block_hash_and_number_prefix)

        block_hashes_list_prefix = rlp_utils.get_length_prefix_list(msg_size)
        msg_size += len(block_hashes_list_prefix)

        msg_bytes = bytearray(msg_size)

        offset = 0

        msg_bytes[offset:len(block_hashes_list_prefix)] = block_hashes_list_prefix
        offset += len(block_hashes_list_prefix)

        msg_bytes[offset:offset + len(block_hash_and_number_prefix)] = block_hash_and_number_prefix
        offset += len(block_hash_and_number_prefix)

        msg_bytes[offset:offset + len(block_hash_prefix)] = block_hash_prefix
        offset += len(block_hash_prefix)

        msg_bytes[offset:offset + len(block_hash_bytes)] = block_hash_bytes
        offset += len(block_hash_bytes)

        msg_bytes[offset:offset + len(number_bytes)] = number_bytes

        return cls(msg_bytes)

    def log_level(self) -> LogLevel:
        return LogLevel.DEBUG

    def prev_block_hash(self) -> Sha256Hash:
        raise NotImplementedError()

    def txns(self):
        raise NotImplementedError()


from bxcommon.utils.object_hash import Sha256Hash
from bxgateway.messages.btc.block_btc_message import BlockBtcMessage
from bxgateway.messages.btc.inventory_btc_message import (
    InvBtcMessage,
    InventoryType,
)
from bxgateway.services.push_block_queuing_service import PushBlockQueuingService


class BtcBlockQueuingService(
    PushBlockQueuingService[BlockBtcMessage, InvBtcMessage]
):
    def build_block_header_message(
        self, block_hash: Sha256Hash, block_message: BlockBtcMessage
    ) -> InvBtcMessage:
        return InvBtcMessage(
            block_message.magic(), [(InventoryType.MSG_BLOCK, block_hash)]
        )

    def get_previous_block_hash_from_message(
        self, block_message: BlockBtcMessage
    ) -> Sha256Hash:
        return block_message.prev_block_hash()

    def on_block_sent(
        self, block_hash: Sha256Hash, block_message: BlockBtcMessage
    ):
        # After sending block message to Bitcoin node sending INV message for the same block to the node
        # This is needed to update Synced Headers value of the gateway peer on the Bitcoin node
        # If Synced Headers is not up-to-date than Bitcoin node does not push compact blocks to the gateway
        inv_msg = InvBtcMessage(
            magic=block_message.magic(),
            inv_vects=[(InventoryType.MSG_BLOCK, block_hash)],
        )
        self.connection.enqueue_msg(inv_msg)


import struct

from bxutils.logging.log_level import LogLevel

from bxcommon.utils import convert
from bxcommon.utils.blockchain_utils.btc.btc_object_hash import BtcObjectHash
from bxgateway import btc_constants
from bxgateway.btc_constants import BTC_HDR_COMMON_OFF
from bxgateway.messages.btc.btc_message import BtcMessage
from bxgateway.messages.btc.btc_message_type import BtcMessageType


class RejectBtcMessage(BtcMessage):
    MESSAGE_TYPE = BtcMessageType.REJECT
    # ccodes
    REJECT_MALFORMED = 0x01
    REJECT_INVALID = 0x10
    REJECT_OBSOLETE = 0x11
    REJECT_DUPLICATE = 0x12
    REJECT_NONSTANDARD = 0x40
    REJECT_DUST = 0x41
    REJECT_INSUFFICIENTFEE = 0x42
    REJECT_CHECKPOINT = 0x43

    def __init__(self, magic=None, message=None, ccode=None, reason=None, b_data=None, buf=None):
        if buf is None:
            buf = bytearray(BTC_HDR_COMMON_OFF + 9 + len(message) + 1 + 9 + len(reason) + len(b_data))
            self.buf = buf

            off = BTC_HDR_COMMON_OFF
            struct.pack_into('<%dpB' % (len(message) + 1,), buf, off, message, ccode)
            off += len(message) + 1 + 1
            struct.pack_into('<%dp' % (len(reason) + 1,), buf, off, reason)
            off += len(reason) + 1
            buf[off:off + len(b_data)] = b_data
            off += len(b_data)

            BtcMessage.__init__(self, magic, self.MESSAGE_TYPE, off - BTC_HDR_COMMON_OFF, buf)
        else:
            self.buf = buf
            self._memoryview = memoryview(buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None

        self._message = self._ccode = self._reason = self._data = self._obj_hash = None

    def message(self):
        if self._message is None:
            off = BTC_HDR_COMMON_OFF
            self._message = struct.unpack_from('%dp' % (len(self.buf) - off,), self.buf, off)[0]
            off += len(self._message) + 1
            self._ccode = struct.unpack_from('B', self.buf, off)[0]
            off += 1
            self._reason = struct.unpack_from('%dp' % (len(self.buf) - off,), self.buf, off)[0]
            off += len(self._reason) + 1
            self._data = self.buf[off:]

            self._obj_hash = BtcObjectHash(buf=self.buf, offset=len(self.buf) - btc_constants.BTC_SHA_HASH_LEN,
                                           length=btc_constants.BTC_SHA_HASH_LEN)
        return self._message

    def ccode(self):
        if self._message is None:
            self.message()
        return self._ccode

    def reason(self):
        if self._message is None:
            self.message()
        return self._reason

    def data(self):
        if self._message is None:
            self.message()
        return self._data

    def obj_hash(self):
        if self._obj_hash is None:
            self.message()
        return self._obj_hash

    def log_level(self):
        return LogLevel.DEBUG

    def __repr__(self):
        return "RejectBtcMessage<message: {}, ccode: {}, reason: {}, obj hash {}".format(
            self.message(), self.ccode(), self.reason(), convert.bytes_to_hex(self.obj_hash().binary))



import gc
import dataclasses
import functools

from dataclasses import dataclass
from collections import defaultdict
from typing import Dict
from typing import Type, Any, TYPE_CHECKING

from bxcommon import constants
from bxcommon.utils.stats.statistics_service import StatsIntervalData, StatisticsService
from bxutils import logging
from bxutils.logging import LogRecordType

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxcommon.connections.abstract_node import AbstractNode


@dataclass
class NodeTransactionStatInterval(StatsIntervalData):
    total_uncollectable: int = 0
    generation_zero_size: int = 0
    generation_one_size: int = 0
    generation_two_size: int = 0
    time_spent_in_gc: float = 0
    collection_counts: Dict[int, int] = dataclasses.field(default_factory=functools.partial(defaultdict, int))
    generation_zero_collections: int = 0
    generation_one_collections: int = 0
    generation_two_collections: int = 0


class _NodeStatisticsService(StatisticsService[NodeTransactionStatInterval, "AbstractNode"]):
    def __init__(self, interval: int = constants.NODE_STATS_INTERVAL_S) -> None:
        super().__init__(
            "NodeStatus",
            interval,
            reset=True,
            stat_logger=logging.get_logger(LogRecordType.NodeStatus, __name__),
        )

    def get_interval_data_class(self) -> Type[NodeTransactionStatInterval]:
        return NodeTransactionStatInterval

    def get_info(self) -> Dict[str, Any]:
        interval_data = self.interval_data
        assert interval_data is not None
        gen0, gen1, gen2 = gc.get_count()
        return {
            "garbage_collection": {
                "uncollectable": len(gc.garbage),
                "collection_counts": {
                    f"gen{k}": v for k, v in interval_data.collection_counts.items()
                },
                "sizes": {"gen0": gen0, "gen1": gen1, "gen2": gen2,},
                "total_elapsed_time": interval_data.time_spent_in_gc,
            }
        }

    def log_gc_duration(self, generation: int, duration_s: float) -> None:
        self.interval_data.time_spent_in_gc += duration_s
        self.interval_data.collection_counts[generation] += 1


node_stats_service = _NodeStatisticsService()



from typing import NamedTuple, Optional


class SSLFileInfo(NamedTuple):
    sub_directory_name: str
    file_name: str
    url: Optional[str] = None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} <" \
            f"sub_directory_name = {self.sub_directory_name}, " \
            f"file_name = {self.file_name}, " \
            f"url = {self.url}" \
            f">"



from asyncio import Protocol
from typing import Optional

from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxcommon.network.ip_endpoint import IpEndpoint


class SocketConnectionProtocolPy36(AbstractSocketConnectionProtocol, Protocol):
    def __init__(
        self,
        # pyre-fixme[11]: Annotation `AbstractNode` is not defined as a type.
        node: "AbstractNode",
        endpoint: Optional[IpEndpoint] = None,
        is_ssl: bool = True,
    ):
        AbstractSocketConnectionProtocol.__init__(self, node, endpoint, is_ssl)

    def data_received(self, data: bytes) -> None:
        if self.is_receivable():
            self._node.on_bytes_received(self.file_no, data)

    def get_last_read_duration_ms(self) -> float:
        return 0

    def get_time_since_read_end_ms(self, end_time: float) -> float:
        return 0



import json
from typing import Dict, Any, Optional


class MockRpcRequest:

    def __init__(self, headers: Dict[str, Any], json_body: str, text: Optional[str] = None) -> None:
        self.headers = headers
        self.raw_headers = headers
        self.remote = "test"
        self._json_body = json_body
        self.raw_path = "/"
        if text is None:
            text = json_body
        self._text = text

    async def text(self) -> str:
        assert self._text is not None
        # pyre-fixme[7]: Expected `str` but got `Optional[str]`.
        return self._text

    async def json(self) -> Dict[str, Any]:
        return json.loads(self._json_body)



import struct

from bxgateway.btc_constants import BTC_HDR_COMMON_OFF
from bxgateway.messages.btc.btc_message import BtcMessage
from bxgateway.messages.btc.btc_message_type import BtcMessageType
from bxgateway.messages.btc.btc_messages_util import ipaddrport_to_btcbytearray, pack_int_to_btc_varint


# the addr argument should be an array of (timestamp, ipaddr, port) triples
class AddrBtcMessage(BtcMessage):
    MESSAGE_TYPE = BtcMessageType.ADDRESS

    def __init__(self, magic=None, addrs=None, buf=None):
        if addrs is None:
            addrs = []
        if buf is None:
            buf = bytearray(BTC_HDR_COMMON_OFF + 9 + len(addrs) * (4 + 18))
            self.buf = buf

            off = BTC_HDR_COMMON_OFF
            off += pack_int_to_btc_varint(len(addrs), buf, off)

            for triplet in addrs:
                # pack the timestamp
                struct.pack_into('<L', buf, off, triplet[0])
                off += 4
                # pack the host ip and port pair
                buf[off:off + 18] = ipaddrport_to_btcbytearray(triplet[1], triplet[2])
                off += 18

            BtcMessage.__init__(self, magic, self.MESSAGE_TYPE, off - BTC_HDR_COMMON_OFF, buf)
        else:
            self.buf = buf
            self._memoryview = memoryview(buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None

    def __iter__(self):
        raise RuntimeError('FIXME')
        # FIXME buf is not defined, change to self.buf and test
        # off = BTC_HDR_COMMON_OFF
        # count, size = btcvarint_to_int(buf, off)
        # off += size
        #
        # for i in range(count):
        #     timestamp = struct.unpack_from('<L', self.buf, off)
        #     off += 4
        #     host, port = btcbytearray_to_ipaddrport(buf[off:off+18])
        #     off += 18
        #     yield (timestamp, host, port)



import struct
from typing import Optional

from bxcommon.constants import UL_INT_SIZE_IN_BYTES, CONTROL_FLAGS_LEN
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxutils.logging.log_level import LogLevel


class TxServiceSyncCompleteMessage(AbstractBloxrouteMessage):
    MESSAGE_TYPE = BloxrouteMessageType.TX_SERVICE_SYNC_COMPLETE

    """
    Message is sent after complete synchronize all txs in a network
    """

    def __init__(self, network_num: Optional[int] = None, buf: Optional[bytearray] = None) -> None:
        if buf is None:
            buf = bytearray(self.HEADER_LENGTH + UL_INT_SIZE_IN_BYTES + CONTROL_FLAGS_LEN)

        self.buf: bytearray = buf

        off = self.HEADER_LENGTH
        if network_num is not None:
            struct.pack_into("<L", self.buf, off, network_num)

        self._network_num: Optional[int] = network_num

        super(TxServiceSyncCompleteMessage, self).__init__(
            self.MESSAGE_TYPE,
            len(self.buf) - self.HEADER_LENGTH,
            self.buf
        )

    def log_level(self) -> LogLevel:
        return LogLevel.DEBUG

    def network_num(self) -> int:
        if self._network_num is None:
            off = self.HEADER_LENGTH
            self._network_num, = struct.unpack_from("<L", self._memoryview, off)

        network_num = self._network_num
        assert network_num is not None
        return network_num

    def __repr__(self) -> str:
        return "{}<network_num: {}".format(self.__class__.__name__, self.network_num())



import typing
from abc import abstractmethod
from typing import List

from bxcommon.connections.connection_type import ConnectionType
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.utils.object_hash import Sha256Hash
from bxgateway import btc_constants, gateway_constants
from bxgateway.connections.abstract_blockchain_connection_protocol import AbstractBlockchainConnectionProtocol
from bxgateway.messages.btc.addr_btc_message import AddrBtcMessage
from bxgateway.messages.btc.btc_message_factory import btc_message_factory
from bxgateway.messages.btc.btc_message_type import BtcMessageType
from bxgateway.messages.btc.inventory_btc_message import InvBtcMessage, InventoryType
from bxgateway.messages.btc.pong_btc_message import PongBtcMessage
from bxgateway.messages.btc.version_btc_message import VersionBtcMessage
from bxutils import logging

logger = logging.get_logger(__name__)


class BtcBaseConnectionProtocol(AbstractBlockchainConnectionProtocol):
    def __init__(self, connection):
        super(BtcBaseConnectionProtocol, self).__init__(connection)
        self.node = typing.cast("bxgateway.connections.btc.btc_gateway_node.BtcGatewayNode", connection.node)
        self.magic = self.node.opts.blockchain_net_magic
        self.version = self.node.opts.blockchain_version

        connection.hello_messages = btc_constants.BTC_HELLO_MESSAGES
        connection.header_size = btc_constants.BTC_HDR_COMMON_OFF
        connection.message_handlers = {
            BtcMessageType.PING: self.msg_ping,
            BtcMessageType.PONG: self.msg_pong,
            BtcMessageType.GET_ADDRESS: self.msg_getaddr
        }

        # Establish connection with blockchain node
        version_msg = VersionBtcMessage(self.magic, self.version, connection.peer_ip, connection.peer_port,
                                        self.node.opts.external_ip, self.node.opts.external_port,
                                        self.node.opts.blockchain_nonce, 0,
                                        f"{gateway_constants.GATEWAY_PEER_NAME} {self.node.opts.source_version}".encode("utf-8"),
                                        self.node.opts.blockchain_services)
        connection.enqueue_msg(version_msg)

    def msg_block(self, msg) -> None:
        """
        Handle block message
        """
        block_hash = msg.block_hash()
        if not self.node.should_process_block_hash(block_hash):
            return
        if self.node.block_cleanup_service.is_marked_for_cleanup(block_hash):
            self.connection.log_trace("Marked block for cleanup: {}", block_hash)
            self.node.block_cleanup_service.clean_block_transactions(
                transaction_service=self.node.get_tx_service(),
                block_msg=msg
            )
        else:
            self.process_msg_block(msg)

        # After receiving block message sending INV message for the same block to Bitcoin node
        # This is needed to update Synced Headers value of the gateway peer on the Bitcoin node
        # If Synced Headers is not up-to-date than Bitcoin node does not push compact blocks to the gateway
        inv_msg = InvBtcMessage(magic=self.node.opts.blockchain_net_magic,
                                inv_vects=[(InventoryType.MSG_BLOCK, msg.block_hash())])
        self.connection.enqueue_msg(inv_msg)

    def msg_ping(self, msg):
        """
        Handle ping messages. Respond with a pong.
        """
        reply = PongBtcMessage(self.magic, msg.nonce())
        self.connection.enqueue_msg(reply)

    def msg_pong(self, msg):
        """
        Handle pong messages. For now, don't do anything since we don't ping.
        """
        pass

    def msg_getaddr(self, _msg):
        """
        Handle a getaddr message. Return a blank address to preserve privacy.
        """
        reply = AddrBtcMessage(self.magic)
        self.connection.enqueue_msg(reply)

    @abstractmethod
    def _build_get_blocks_message_for_block_confirmation(self, hashes: List[Sha256Hash]) -> AbstractMessage:
        pass



from typing import Optional, Union

from bxcommon.network.network_direction import NetworkDirection
from bxcommon.utils.stats.measurement_type import MeasurementType
from bxcommon.utils.stats.memory_statistics_service import memory_statistics
from bxcommon.utils.stats.throughput_service import throughput_statistics


def add_throughput_event(
    direction: NetworkDirection,
    msg_type: str,
    msg_size: int,
    peer_desc: str,
    peer_id: Optional[str] = None,
):
    return throughput_statistics.add_event(direction, msg_type, msg_size, peer_desc, peer_id)


def add_measurement(
    peer_desc: str,
    measure_type: MeasurementType,
    measure_value: Union[float, int],
    peer_id: Optional[str] = None,
):
    return throughput_statistics.add_measurement(peer_desc, measure_type, measure_value, peer_id)


def add_obj_mem_stats(
    class_name,
    network_num,
    obj,
    obj_name,
    obj_mem_info,
    object_type=None,
    size_type=None,
    object_item_count=None,
):
    return memory_statistics.add_mem_stats(
        class_name,
        network_num,
        obj,
        obj_name,
        obj_mem_info,
        object_type,
        size_type,
        object_item_count,
    )


def increment_obj_mem_stats(
    class_name,
    network_num,
    obj,
    obj_name,
    obj_mem_info,
    object_type=None,
    size_type=None,
    object_item_count=None,
):
    return memory_statistics.increment_mem_stats(
        class_name,
        network_num,
        obj,
        obj_name,
        obj_mem_info,
        object_type,
        size_type,
        object_item_count,
    )


def reset_class_mem_stats(class_name):
    return memory_statistics.reset_class_mem_stats(class_name)



import time
from collections import defaultdict
from enum import Enum
from typing import Dict, Set, List, NamedTuple

from bxcommon.utils import crypto
from bxcommon.utils.alarm_queue import AlarmQueue
from bxcommon.utils.expiration_queue import ExpirationQueue
from bxcommon.utils.object_hash import Sha256Hash
from bxgateway import gateway_constants
from bxutils import logging

logger = logging.get_logger(__name__)


class BlockRecoveryInfo(NamedTuple):
    block_hash: Sha256Hash
    unknown_short_ids: Set[int]
    unknown_transaction_hashes: Set[Sha256Hash]
    recovery_start_time: float


class RecoveredTxsSource(Enum):
    TXS_RECEIVED_FROM_BDN = "TransactionsReceivedFromBdn"
    TXS_RECEIVED_FROM_NODE = "TransactionsReceivedFromNode"
    TXS_RECOVERED = "TxsRecovered"
    COMPRESSED_BLOCK_TXS_RECEIVED = "CompressedBlockTxsReceived"
    BLOCK_RECEIVED_FROM_NODE = "BlockReceivedFromNode"

    def __str__(self):
        return self.name


class BlockRecoveryService:
    """
    Service class that handles blocks gateway receives with unknown transaction short ids are contents.

    Attributes
    ----------
    recovered blocks: queue to which recovered blocks are pushed to
    _alarm_queue: reference to alarm queue to schedule cleanup on

    _bx_block_hash_to_sids: map of compressed block hash to its set of unknown short ids
    _bx_block_hash_to_tx_hashes: map of compressed block hash to its set of unknown transaction hashes
    _bx_block_hash_to_block_hash: map of compressed block hash to its original block hash
    _bx_block_hash_to_block: map of compressed block hash to its compressed byte representation
    _block_hash_to_bx_block_hashes: map of original block hash to compressed block hashes waiting for recovery
    _sid_to_bx_block_hashes: map of short id to compressed block hashes waiting for recovery
    _tx_hash_to_bx_block_hashes: map of transaction hash to block hashes waiting for recovery

    _cleanup_scheduled: whether block recovery has an alarm scheduled to clean up recovering blocks
    _blocks_expiration_queue: queue to trigger expiration of waiting for block recovery
    """

    _alarm_queue: AlarmQueue
    _bx_block_hash_to_sids: Dict[Sha256Hash, Set[int]]
    _bx_block_hash_to_tx_hashes: Dict[Sha256Hash, Set[Sha256Hash]]
    _bx_block_hash_to_block_hash: Dict[Sha256Hash, Sha256Hash]
    _bx_block_hash_to_block: Dict[Sha256Hash, memoryview]
    _block_hash_to_bx_block_hashes: Dict[Sha256Hash, Set[Sha256Hash]]
    _sid_to_bx_block_hashes: Dict[int, Set[Sha256Hash]]
    _tx_hash_to_bx_block_hashes: Dict[Sha256Hash, Set[Sha256Hash]]
    _blocks_expiration_queue: ExpirationQueue
    _cleanup_scheduled: bool = False

    recovery_attempts_by_block: Dict[Sha256Hash, int]

    def __init__(self, alarm_queue: AlarmQueue):
        self.recovered_blocks = []

        self._alarm_queue = alarm_queue

        self._bx_block_hash_to_sids = {}
        self._bx_block_hash_to_tx_hashes = {}
        self._bx_block_hash_to_block_hash = {}
        self._bx_block_hash_to_block = {}
        self.recovery_attempts_by_block = defaultdict(int)
        self._block_hash_to_bx_block_hashes = defaultdict(set)
        self._sid_to_bx_block_hashes = defaultdict(set)
        self._tx_hash_to_bx_block_hashes: Dict[Sha256Hash, Set[Sha256Hash]] = defaultdict(set)
        self._blocks_expiration_queue = ExpirationQueue(gateway_constants.BLOCK_RECOVERY_MAX_QUEUE_TIME)

    def add_block(self, bx_block: memoryview, block_hash: Sha256Hash, unknown_tx_sids: List[int],
                  unknown_tx_hashes: List[Sha256Hash]):
        """
        Adds a block that needs to recovery. Tracks unknown short ids and contents as they come in.
        :param bx_block: bytearray representation of compressed block
        :param block_hash: original ObjectHash of block
        :param unknown_tx_sids: list of unknown short ids
        :param unknown_tx_hashes: list of unknown tx ObjectHashes
        """
        logger.trace("Recovering block with {} unknown short ids and {} contents: {}", len(unknown_tx_sids),
                     len(unknown_tx_hashes), block_hash)
        bx_block_hash = Sha256Hash(crypto.double_sha256(bx_block))

        self._bx_block_hash_to_block[bx_block_hash] = bx_block
        self._bx_block_hash_to_block_hash[bx_block_hash] = block_hash
        self._bx_block_hash_to_sids[bx_block_hash] = set(unknown_tx_sids)
        self._bx_block_hash_to_tx_hashes[bx_block_hash] = set(unknown_tx_hashes)

        self._block_hash_to_bx_block_hashes[block_hash].add(bx_block_hash)
        for sid in unknown_tx_sids:
            self._sid_to_bx_block_hashes[sid].add(bx_block_hash)
        for tx_hash in unknown_tx_hashes:
            self._tx_hash_to_bx_block_hashes[tx_hash].add(bx_block_hash)

        self._blocks_expiration_queue.add(bx_block_hash)
        self._schedule_cleanup()

    def get_blocks_awaiting_recovery(self) -> List[BlockRecoveryInfo]:
        """
        Fetch all blocks still awaiting recovery and retry.
        """
        blocks_awaiting_recovery = []
        for block_hash, bx_block_hashes in self._block_hash_to_bx_block_hashes.items():
            unknown_short_ids = set()
            unknown_transaction_hashes = set()
            for bx_block_hash in bx_block_hashes:
                unknown_short_ids.update(self._bx_block_hash_to_sids[bx_block_hash])
                unknown_transaction_hashes.update(self._bx_block_hash_to_tx_hashes[bx_block_hash])
            blocks_awaiting_recovery.append(BlockRecoveryInfo(block_hash, unknown_short_ids, unknown_transaction_hashes,
                                                              time.time()))
        return blocks_awaiting_recovery

    def check_missing_sid(self, sid: int, recovered_txs_source: RecoveredTxsSource) -> bool:
        """
        Resolves recovering blocks depend on sid.
        :param sid: SID info that has been processed
        :param recovered_txs_source: source of recovered transaction
        """
        if sid in self._sid_to_bx_block_hashes:
            logger.trace("Resolved previously unknown short id: {0}.", sid)

            bx_block_hashes = self._sid_to_bx_block_hashes[sid]
            for bx_block_hash in bx_block_hashes:
                if bx_block_hash in self._bx_block_hash_to_sids:
                    if sid in self._bx_block_hash_to_sids[bx_block_hash]:
                        self._bx_block_hash_to_sids[bx_block_hash].discard(sid)
                        self._check_if_recovered(bx_block_hash, recovered_txs_source)

            del self._sid_to_bx_block_hashes[sid]
            return True
        else:
            return False

    def check_missing_tx_hash(self, tx_hash: Sha256Hash, recovered_txs_source: RecoveredTxsSource) -> bool:
        """
        Resolves recovering blocks depend on transaction hash.
        :param tx_hash: transaction info that has been processed
        :param recovered_txs_source: source of recovered transaction
        """
        if tx_hash in self._tx_hash_to_bx_block_hashes:
            logger.trace("Resolved previously unknown transaction hash {0}.", tx_hash)

            bx_block_hashes = self._tx_hash_to_bx_block_hashes[tx_hash]
            for bx_block_hash in bx_block_hashes:
                if bx_block_hash in self._bx_block_hash_to_tx_hashes:
                    if tx_hash in self._bx_block_hash_to_tx_hashes[bx_block_hash]:
                        self._bx_block_hash_to_tx_hashes[bx_block_hash].discard(tx_hash)
                        self._check_if_recovered(bx_block_hash, recovered_txs_source)

            del self._tx_hash_to_bx_block_hashes[tx_hash]
            return True
        else:
            return False

    def cancel_recovery_for_block(self, block_hash: Sha256Hash) -> bool:
        """
        Cancels recovery for all compressed blocks matching a block hash
        :param block_hash: ObjectHash
        """
        if block_hash in self._block_hash_to_bx_block_hashes:
            logger.trace("Cancelled block recovery for block: {}", block_hash)
            self._remove_recovered_block_hash(block_hash)
            return True
        else:
            return False

    def awaiting_recovery(self, block_hash: Sha256Hash) -> bool:
        return block_hash in self._block_hash_to_bx_block_hashes

    # pyre-fixme[9]: clean_up_time has type `float`; used as `None`.
    def cleanup_old_blocks(self, clean_up_time: float = None):
        """
        Cleans up old compressed blocks awaiting recovery.
        :param clean_up_time:
        """
        logger.debug("Cleaning up block recovery.")
        num_blocks_awaiting_recovery = len(self._bx_block_hash_to_block)
        self._blocks_expiration_queue.remove_expired(current_time=clean_up_time,
                                                     remove_callback=self._remove_not_recovered_block)
        logger.debug("Cleaned up {} blocks awaiting recovery.",
                     num_blocks_awaiting_recovery - len(self._bx_block_hash_to_block))

        if self._bx_block_hash_to_block:
            return gateway_constants.BLOCK_RECOVERY_MAX_QUEUE_TIME

        # disable clean up until receive the next block with unknown tx
        self._cleanup_scheduled = False
        return 0

    def clean_up_recovered_blocks(self):
        """
        Cleans up blocks that have finished recovery.
        :return:
        """
        logger.trace("Cleaning up {} recovered blocks.", len(self.recovered_blocks))
        del self.recovered_blocks[:]

    def _check_if_recovered(self, bx_block_hash: Sha256Hash, recovered_txs_source: RecoveredTxsSource):
        """
        Checks if a compressed block has received all short ids and transaction hashes necessary to recover.
        Adds block to recovered blocks if so.
        :param bx_block_hash: ObjectHash
        :return:
        """
        if self._is_block_recovered(bx_block_hash):
            bx_block = self._bx_block_hash_to_block[bx_block_hash]
            block_hash = self._bx_block_hash_to_block_hash[bx_block_hash]
            logger.debug(
                "Recovery status for block {}, compress block hash {}: "
                "Block recovered by gateway. Source of recovered txs is {}.",
                block_hash, bx_block_hash, recovered_txs_source
            )
            self._remove_recovered_block_hash(block_hash)
            self.recovered_blocks.append((bx_block, recovered_txs_source))

    def _is_block_recovered(self, bx_block_hash: Sha256Hash):
        """
        Indicates if a compressed block has received all short ids and transaction hashes necessary to recover.
        :param bx_block_hash: ObjectHash
        :return:
        """
        return len(self._bx_block_hash_to_sids[bx_block_hash]) == 0 and len(
            self._bx_block_hash_to_tx_hashes[bx_block_hash]) == 0

    def _remove_recovered_block_hash(self, block_hash: Sha256Hash):
        """
        Removes all compressed blocks awaiting recovery that are matches to a recovered block hash.
        :param block_hash: ObjectHash
        :return:
        """
        if block_hash in self._block_hash_to_bx_block_hashes:
            for bx_block_hash in self._block_hash_to_bx_block_hashes[block_hash]:
                if bx_block_hash in self._bx_block_hash_to_block:
                    self._remove_sid_and_tx_mapping_for_bx_block_hash(bx_block_hash)
                    del self._bx_block_hash_to_block[bx_block_hash]
                    del self._bx_block_hash_to_block_hash[bx_block_hash]
            del self._block_hash_to_bx_block_hashes[block_hash]

    def _remove_sid_and_tx_mapping_for_bx_block_hash(self, bx_block_hash: Sha256Hash):
        """
        Removes all short id and transaction mapping for a compressed block.
        :param bx_block_hash:
        :return:
        """
        if bx_block_hash not in self._bx_block_hash_to_block:
            raise ValueError("Can't remove mapping for a block that isn't being recovered.")

        for sid in self._bx_block_hash_to_sids[bx_block_hash]:
            if sid in self._sid_to_bx_block_hashes:
                self._sid_to_bx_block_hashes[sid].discard(bx_block_hash)
                if len(self._sid_to_bx_block_hashes[sid]) == 0:
                    del self._sid_to_bx_block_hashes[sid]
        del self._bx_block_hash_to_sids[bx_block_hash]

        for tx_hash in self._bx_block_hash_to_tx_hashes[bx_block_hash]:
            if tx_hash in self._tx_hash_to_bx_block_hashes:
                self._tx_hash_to_bx_block_hashes[tx_hash].discard(bx_block_hash)
                if len(self._tx_hash_to_bx_block_hashes[tx_hash]) == 0:
                    del self._tx_hash_to_bx_block_hashes[tx_hash]
        del self._bx_block_hash_to_tx_hashes[bx_block_hash]

    def _remove_not_recovered_block(self, bx_block_hash: Sha256Hash):
        """
        Removes compressed block that has not recovered.
        :param bx_block_hash: ObjectHash
        """
        if bx_block_hash in self._bx_block_hash_to_block:
            logger.trace("Block has failed recovery: {}", bx_block_hash)

            self._remove_sid_and_tx_mapping_for_bx_block_hash(bx_block_hash)
            del self._bx_block_hash_to_block[bx_block_hash]

            block_hash = self._bx_block_hash_to_block_hash.pop(bx_block_hash)
            self._block_hash_to_bx_block_hashes[block_hash].discard(bx_block_hash)
            if len(self._block_hash_to_bx_block_hashes[block_hash]) == 0:
                del self._block_hash_to_bx_block_hashes[block_hash]

    def _schedule_cleanup(self):
        if not self._cleanup_scheduled and self._bx_block_hash_to_block:
            logger.trace("Scheduling block recovery cleanup in {} seconds.",
                         gateway_constants.BLOCK_RECOVERY_MAX_QUEUE_TIME)
            self._alarm_queue.register_alarm(gateway_constants.BLOCK_RECOVERY_MAX_QUEUE_TIME, self.cleanup_old_blocks)
            self._cleanup_scheduled = True



from bxcommon.models.serializable_flag import SerializableFlag


class QuotaType(SerializableFlag):
    FREE_DAILY_QUOTA = 1
    PAID_DAILY_QUOTA = 2

    def __str__(self):
        return str(self.name).lower()[:4]



from typing import Optional, Set, Union

from bxcommon.connections.abstract_connection import AbstractConnection
from bxcommon.connections.connection_state import ConnectionState
from bxcommon.connections.connection_type import ConnectionType
from bxcommon.constants import PING_INTERVAL_S
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.messages.abstract_message_factory import AbstractMessageFactory
from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxcommon.network.network_direction import NetworkDirection
from bxcommon.utils import memory_utils
from bxcommon.utils.buffers.input_buffer import InputBuffer
from bxcommon.utils.buffers.output_buffer import OutputBuffer
from bxcommon.utils.memory_utils import SpecialMemoryProperties, SpecialTuple
from bxcommon.models.broadcast_message_type import BroadcastMessageType


class MockConnection(AbstractConnection, SpecialMemoryProperties):
    CONNECTION_TYPE = ConnectionType.EXTERNAL_GATEWAY

    # pylint: disable=super-init-not-called
    def __init__(self, sock: AbstractSocketConnectionProtocol, node) -> None:
        self.socket_connection = sock
        self.file_no = sock.file_no

        # (IP, Port) at time of socket creation. We may get a new application level port in
        # the version message if the connection is not from me.
        self.peer_ip, self.peer_port = sock.endpoint
        self.endpoint = sock.endpoint
        self.peer_id = None
        self.my_ip = node.opts.external_ip
        self.my_port = node.opts.external_port
        self.direction = self.socket_connection.direction

        self.from_me = self.direction == NetworkDirection.OUTBOUND  # Whether or not I initiated the connection

        self.outputbuf = OutputBuffer()
        self.inputbuf = InputBuffer()
        self.node = node

        self.is_persistent = False
        self.state = ConnectionState.CONNECTING
        self.established = False

        # Number of bad messages I've received in a row.
        self.num_bad_messages = 0
        self.peer_desc = "{} {}".format(self.peer_ip, self.peer_port)
        self.message_handlers = None
        self.network_num = node.opts.blockchain_network_num
        self.format_connection()

        self.enqueued_messages = []
        self.node_privileges = "general"
        self.subscribed_broadcasts = [BroadcastMessageType.BLOCK]
        self.ping_alarm_id = None

    def __repr__(self):
        return f"MockConnection<file_no: {self.file_no}, address: ({self.peer_ip}, {self.peer_port}), " \
            f"network_num: {self.network_num}>"

    def connection_message_factory(self) -> AbstractMessageFactory:
        pass

    def ping_message(self) -> AbstractMessage:
        pass

    def add_received_bytes(self, bytes_received):
        self.inputbuf.add_bytes(bytes_received)
        self.mark_for_close()

    def get_bytes_to_send(self):
        return self.outputbuf.output_msgs[0]

    def advance_sent_bytes(self, bytes_sent):
        self.advance_bytes_on_buffer(self.outputbuf, bytes_sent)

    def advance_bytes_on_buffer(self, buf, bytes_written):
        buf.advance_buffer(bytes_written)

    def enqueue_msg(self, msg: AbstractMessage, prepend: bool = False):
        if not self.is_alive():
            return

        self.outputbuf.enqueue_msgbytes(msg.rawbytes())
        self.enqueued_messages.append(msg)

    def enqueue_msg_bytes(
        self,
        msg_bytes: Union[bytearray, memoryview],
        prepend: bool = False,
    ):
        if not self.is_alive():
            return

        self.outputbuf.enqueue_msgbytes(msg_bytes)
        self.enqueued_messages.append(msg_bytes)

    def process_message(self):
        pass

    def send_ping(self):
        return PING_INTERVAL_S

    def special_memory_size(self, ids: Optional[Set[int]] = None) -> SpecialTuple:
        return memory_utils.add_special_objects(self.inputbuf, self.outputbuf, ids=ids)



import asyncio
from typing import TYPE_CHECKING

from bxcommon.rpc import rpc_constants
from bxcommon.rpc.bx_json_rpc_request import BxJsonRpcRequest
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.requests.abstract_rpc_request import AbstractRpcRequest
from bxgateway.rpc.gateway_status_details_level import GatewayStatusDetailsLevel
from bxgateway.utils.logging.status import status_log
from bxutils.encoding.json_encoder import EnhancedJSONEncoder

if TYPE_CHECKING:
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode


class GatewayStatusRpcRequest(AbstractRpcRequest["AbstractGatewayNode"]):
    help = {
        "params": f"Optional - {rpc_constants.DETAILS_LEVEL_PARAMS_KEY}: "
                  f"{list(GatewayStatusDetailsLevel.__members__.keys())}.",
        "description": "return status of the bloXroute Gateway"
    }

    def __init__(
        self,
        request: BxJsonRpcRequest,
        node: "AbstractGatewayNode",
    ):
        if request.params is None:
            request.params = {
                rpc_constants.DETAILS_LEVEL_PARAMS_KEY: GatewayStatusDetailsLevel.SUMMARY.name
            }
        self._details_level = None
        super().__init__(request, node)
        self._json_encoder = EnhancedJSONEncoder()
        params = request.params
        assert isinstance(params, dict)
        self.params = params

    def validate_params(self) -> None:
        super().validate_params()
        params = self.params
        assert isinstance(params, dict)
        if rpc_constants.DETAILS_LEVEL_PARAMS_KEY not in params:
            params[rpc_constants.DETAILS_LEVEL_PARAMS_KEY] = GatewayStatusDetailsLevel.SUMMARY.name

        self._details_level = params[rpc_constants.DETAILS_LEVEL_PARAMS_KEY].lower()

    async def process_request(self) -> JsonRpcResponse:
        loop = asyncio.get_event_loop()
        opts = self.node.opts
        diagnostics = await loop.run_in_executor(
            self.node.requester.thread_pool,
            status_log.get_diagnostics,
            opts.use_extensions,
            opts.source_version,
            opts.external_ip,
            opts.continent,
            opts.country,
            opts.should_update_source_version,
            self.node.account_id,
            self.node.quota_level
        )
        if self._details_level == GatewayStatusDetailsLevel.SUMMARY.name.lower():
            data = diagnostics.summary
        else:
            data = diagnostics
        result = self._json_encoder.as_dict(data)
        return self.ok(result)



import struct
from typing import NamedTuple, Optional, List, Union

from bxcommon import constants
from bxcommon.models.transaction_flag import TransactionFlag
from bxcommon.utils.crypto import SHA256_HASH_LEN
from bxcommon.utils.object_hash import Sha256Hash
from bxutils import logging

logger = logging.get_logger(__name__)


class TxContentShortIds(NamedTuple):
    tx_hash: Sha256Hash
    tx_content: Optional[Union[bytearray, memoryview]]
    short_ids: List[int]
    short_id_flags: List[TransactionFlag]


def get_serialized_tx_content_short_ids_bytes_len(tx_content_short_ids: TxContentShortIds) -> int:
    """
    Calculates length of tx content and short ids serialized into bytes

    :param tx_content_short_ids: transaction content and short ids
    :return: length of serialized bytes = tx_hash_size + tx_content_size + len(tx_content)
             + expiration_time_size + short_ids_count + short_ids_count * short_id_size
    """
    short_ids_count = 0
    if tx_content_short_ids.short_ids is not None:
        short_ids_count = len(tx_content_short_ids.short_ids)

    tx_content_size = 0
    if tx_content_short_ids.tx_content is not None:
        # pyre-fixme[6]: Expected `Sized` for 1st param but got
        #  `Optional[Union[bytearray, memoryview]]`.
        tx_content_size = len(tx_content_short_ids.tx_content)

    return SHA256_HASH_LEN + constants.UL_INT_SIZE_IN_BYTES + tx_content_size + constants.UL_INT_SIZE_IN_BYTES + \
        constants.UL_SHORT_SIZE_IN_BYTES + (constants.SID_LEN + constants.TRANSACTION_FLAG_LEN) * short_ids_count


def get_serialized_txs_content_short_ids_bytes_len(txs_content_short_ids: List[TxContentShortIds]) -> int:
    return sum([get_serialized_tx_content_short_ids_bytes_len(tx) for tx in txs_content_short_ids])


def serialize_txs_content_short_ids_into_bytes(
    txs_content_short_ids: List[TxContentShortIds], network_num: int
) -> bytearray:
    """
    Serializes list of txs content and short ids into bytes
    :param txs_content_short_ids: list of tuple txs content and short ids
    :param network_num: tx's network number
    :return: bytearray with serialized bytes
    """

    buffer = bytearray(get_serialized_txs_content_short_ids_bytes_len(txs_content_short_ids))
    off = 0
    for tx_content_short_ids in txs_content_short_ids:
        if tx_content_short_ids.tx_content is not None and constants.NULL_TX_SID not in tx_content_short_ids.short_ids:
            buffer[off: off + SHA256_HASH_LEN] = tx_content_short_ids.tx_hash.binary
            off += SHA256_HASH_LEN

            # pyre-fixme[6]: Expected `Sized` for 1st param but got
            #  `Optional[Union[bytearray, memoryview]]`.
            struct.pack_into("<L", buffer, off, len(tx_content_short_ids.tx_content))
            off += constants.UL_INT_SIZE_IN_BYTES
            # pyre-fixme[6]: Expected `Sized` for 1st param but got
            #  `Optional[Union[bytearray, memoryview]]`.
            # pyre-fixme[6]: Expected `Union[typing.Iterable[int], bytes]` for 2nd
            #  param but got `Optional[Union[bytearray, memoryview]]`.
            buffer[off: off + len(tx_content_short_ids.tx_content)] = tx_content_short_ids.tx_content
            # pyre-fixme[6]: Expected `Sized` for 1st param but got
            #  `Optional[Union[bytearray, memoryview]]`.
            off += len(tx_content_short_ids.tx_content)

            # expiration date
            struct.pack_into("<L", buffer, off, 0)
            off += constants.UL_INT_SIZE_IN_BYTES

            struct.pack_into("<H", buffer, off, len(tx_content_short_ids.short_ids))
            off += constants.UL_SHORT_SIZE_IN_BYTES

            for short_id in tx_content_short_ids.short_ids:
                struct.pack_into("<L", buffer, off, short_id)
                off += constants.SID_LEN

            # we pass a transaction flags for each short id provided
            # we pass an array of flags following the array of short ids, in matching order
            # we require that transaction flags array to be the same length of the short_ids array,
            # if validation fail, we will pass empty values as the transaction flags,
            # to make backwards compatibility easier.
            assert len(tx_content_short_ids.short_id_flags) == len(tx_content_short_ids.short_ids),\
                "Invalid TransactionFlag Array Provided"
            for short_id_tx_flag in tx_content_short_ids.short_id_flags:
                struct.pack_into("<H", buffer, off, short_id_tx_flag.value)
                off += constants.TRANSACTION_FLAG_LEN

        else:
            logger.debug(
                "Transaction {} in network {} is missing either content or short ids. short id is None: {}, "
                "tx content is None: {}", tx_content_short_ids.tx_hash, network_num,
                tx_content_short_ids.short_ids is None, tx_content_short_ids.tx_content is None
            )

    return buffer


def deserialize_txs_content_short_ids_from_buffer(
    buffer: Union[bytearray, memoryview], offset: int, tx_count: int
) -> List[TxContentShortIds]:
    """
    Deserialize list of txs content short ids from buffer

    :param buffer: buffer containing serialized txs content short ids bytes
    :param offset: offset in a buffer where serialized txs content short ids begin
    :param tx_count: how many txs to deserialize
    :return: list of txs content short ids
    """

    txs_content_short_ids: List[TxContentShortIds] = []
    for _ in range(tx_count):
        tx_hash = Sha256Hash(buffer[offset:offset + SHA256_HASH_LEN])
        offset = offset + SHA256_HASH_LEN
        tx_content_size, = struct.unpack_from("<L", buffer, offset)
        offset += constants.UL_INT_SIZE_IN_BYTES
        tx_content = buffer[offset: offset + tx_content_size]
        offset += tx_content_size
        offset += constants.UL_INT_SIZE_IN_BYTES
        short_ids_count, = struct.unpack_from("<H", buffer, offset)
        offset += constants.UL_SHORT_SIZE_IN_BYTES

        short_ids = []
        short_id_flags = []
        for _ in range(short_ids_count):
            short_id, = struct.unpack_from("<L", buffer, offset)
            short_ids.append(short_id)
            offset += constants.SID_LEN
        for _ in range(short_ids_count):
            short_id_flag, = struct.unpack_from("<H", buffer, offset)
            short_id_flags.append(TransactionFlag(short_id_flag))
            offset += constants.TRANSACTION_FLAG_LEN
        txs_content_short_ids.append(TxContentShortIds(tx_hash, tx_content, short_ids, short_id_flags))

    return txs_content_short_ids



import dataclasses
import functools
import sys

from dataclasses import dataclass
from collections import defaultdict
from datetime import datetime
from typing import Dict, Type, Any, TYPE_CHECKING

from bxcommon import constants
from bxcommon.utils import memory_utils
from bxcommon.utils.sizer import Sizer
from bxcommon.utils.stats.class_mem_stats import ClassMemStats
from bxcommon.utils.stats.statistics_service import StatsIntervalData, ThreadedStatisticsService
from bxutils import logging
from bxutils.logging.log_record_type import LogRecordType

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxcommon.connections.abstract_node import AbstractNode


@dataclass
class MemoryStatsIntervalData(StatsIntervalData):
    class_mem_stats: Dict[str, ClassMemStats] = dataclasses.field(
        default_factory=functools.partial(defaultdict, ClassMemStats)
    )


# pyre-fixme[24]: Type parameter `AbstractNode` violates constraints on `T` in
#  generic type `ThreadedStatisticsService`.
# pyre-fixme[24]: Type parameter `MemoryStatsIntervalData` violates constraints on
#  `N` in generic type `ThreadedStatisticsService`.
class MemoryStatsService(ThreadedStatisticsService[MemoryStatsIntervalData, "AbstractNode"]):
    def __init__(
        self,
        interval: int = 0
    ) -> None:
        self.sizer_obj = Sizer()
        super(MemoryStatsService, self).__init__(
            "MemoryStats",
            interval=interval,
            look_back=5,
            reset=False,
            stat_logger=logging.get_logger(LogRecordType.Memory, __name__),
        )

    def get_interval_data_class(self) -> Type[MemoryStatsIntervalData]:
        return MemoryStatsIntervalData

    def set_node(self, node: "AbstractNode") -> None:
        super(MemoryStatsService, self).set_node(node)
        self.interval = node.opts.memory_stats_interval
        self.sizer_obj = Sizer(node)

    def add_mem_stats(
        self,
        class_name,
        network_num,
        obj,
        obj_name,
        obj_mem_info,
        object_type=None,
        size_type=None,
        object_item_count=None,
    ):
        # If the object being analyzed doesn't have a length property
        if object_item_count is None:
            object_item_count = len(obj) if hasattr(obj, "__len__") else 0

        if (
            obj_mem_info.size < constants.MEM_STATS_OBJECT_SIZE_THRESHOLD
            and object_item_count < constants.MEM_STATS_OBJECT_COUNT_THRESHOLD
        ):
            return

        mem_stats = self.interval_data.class_mem_stats[class_name]
        mem_stats.timestamp = datetime.utcnow()

        mem_stats.networks[network_num].analyzed_objects[
            obj_name
        ].object_item_count = object_item_count
        mem_stats.networks[network_num].analyzed_objects[obj_name].object_size = obj_mem_info.size
        mem_stats.networks[network_num].analyzed_objects[
            obj_name
        ].object_flat_size = obj_mem_info.flat_size
        mem_stats.networks[network_num].analyzed_objects[
            obj_name
        ].is_actual_size = obj_mem_info.is_actual_size
        mem_stats.networks[network_num].analyzed_objects[obj_name].object_type = object_type
        mem_stats.networks[network_num].analyzed_objects[obj_name].size_type = size_type

    def get_info(self) -> Dict[str, Any]:
        # total_mem_usage is the peak mem usage fo the process (kilobytes on Linux, bytes on OS X)
        node = self.node
        assert node is not None
        payload = {
            "node_id": node.opts.node_id,
            "node_type": node.opts.node_type,
            "node_network_num": node.opts.blockchain_network_num,
            "node_address": f"{node.opts.external_ip}:{node.opts.external_port}",
            "total_mem_usage": memory_utils.get_app_memory_usage(),
            # pyre-ignore having some difficulty with subclassing generics
            "classes": self.interval_data.class_mem_stats,
        }

        return payload

    def flush_info(self, threshold: int = sys.maxsize) -> int:
        node = self.node
        assert node is not None
        total_memory = memory_utils.get_app_memory_usage()
        node.dump_memory_usage(total_memory, threshold)

        return super(MemoryStatsService, self).flush_info(threshold)

    def increment_mem_stats(
        self,
        class_name,
        network_num,
        obj,
        obj_name,
        obj_mem_info,
        object_type=None,
        size_type=None,
        object_item_count=None,
    ):
        mem_stats = self.interval_data.class_mem_stats[class_name]

        # If the object being analyzed doesn't have a length property
        if object_item_count is None:
            object_item_count = len(obj) if hasattr(obj, "__len__") else 1

        mem_stats.networks[network_num].analyzed_objects[
            obj_name
        ].object_item_count += object_item_count
        mem_stats.networks[network_num].analyzed_objects[obj_name].object_size += obj_mem_info.size
        mem_stats.networks[network_num].analyzed_objects[
            obj_name
        ].object_flat_size += obj_mem_info.flat_size
        mem_stats.networks[network_num].analyzed_objects[
            obj_name
        ].is_actual_size = obj_mem_info.is_actual_size
        mem_stats.networks[network_num].analyzed_objects[obj_name].object_type = object_type
        mem_stats.networks[network_num].analyzed_objects[obj_name].size_type = size_type

    def reset_class_mem_stats(self, class_name) -> None:
        mem_stats = ClassMemStats()
        mem_stats.timestamp = datetime.utcnow()
        # pyre-ignore having some difficulty with subclassing generics
        self.interval_data.class_mem_stats[class_name] = mem_stats


memory_statistics = MemoryStatsService(constants.MEMORY_STATS_INTERVAL_S)



from datetime import datetime, date, timedelta

from bxcommon.utils.port_range import PortRange
from bxutils.logging.fluentd_overflow_handler_type import OverflowHandlerType
from bxutils.logging.log_format import LogFormat
from bxutils.logging.log_level import LogLevel

DEFAULT_LOG_LEVEL = LogLevel.INFO
DEFAULT_STATS_LOG_LEVEL = LogLevel.CRITICAL
STATS_LOGGER_NAMES = ["stats", "bx"]

# Log Format from Python3 Logging Package
#
#   %(name)s            Name of the logger (logging channel)
#   %(levelno)s         Numeric logging level for the message (DEBUG, INFO,
#                       WARNING, ERROR, CRITICAL)
#   %(levelname)s       Text logging level for the message ("DEBUG", "INFO",
#                       "WARNING", "ERROR", "CRITICAL")
#   %(pathname)s        Full pathname of the source file where the logging
#                       call was issued (if available)
#   %(filename)s        Filename portion of pathname
#   %(module)s          Module (name portion of filename)
#   %(lineno)d          Source line number where the logging call was issued
#                       (if available)
#   %(funcName)s        Function name
#   %(created)f         Time when the LogRecord was created (time.time()
#                       return value)
#   %(asctime)s         Textual time when the LogRecord was created
#   %(msecs)d           Millisecond portion of the creation time
#   %(relativeCreated)d Time in milliseconds when the LogRecord was created,
#                       relative to the time the logging module was loaded
#                       (typically at application startup time)
#   %(thread)d          Thread ID (if available)
#   %(threadName)s      Thread name (if available)
#   %(process)d         Process ID (if available)
#   %(message)s         The result of record.getMessage(), computed just as
#                       the record is emitted
DEBUG_LOG_FORMAT_PATTERN = "%(asctime)s - %(process)d - %(name)s - %(levelname)s - %(message)s"
INFO_LOG_FORMAT_PATTERN = "%(asctime)s - %(levelname)s - %(message)s"
PLAIN_LOG_DATE_FORMAT_PATTERN = "%Y-%m-%dT%H:%M:%S.%f %z"
DEFAULT_LOG_FORMAT = LogFormat.PLAIN
FLUENTD_HOST = None
FLUENTD_PORT = 24224
FLUENTD_OVERFLOW_HANDLER = OverflowHandlerType.Print
FLUENTD_LOGGER_MAX_QUEUE_SIZE = 2 * 1024 * 1024

FLUENTD_DEFAULT_TAG = "bx"

# ssl constants

# TODO: use the data dir configurations (https://github.com/bloXroute-Labs/bxcommon-private/pull/510) after merging
#  bxutils back into bxcommon
DEFAULT_EXPIRATION_DATE: date = datetime.today().date() + timedelta(365)
SSL_FOLDER = ".ssl"
SSL_KEY_FILE_FORMAT = "{}_key.pem"
SSL_CERT_FILE_FORMAT = "{}_cert.pem"
DEFAULT_PUBLIC_CA_URL = "https://s3.amazonaws.com/credentials.blxrbdn.com/"
DEFAULT_CERTIFICATE_RENEWAL_PERIOD_DAYS: int = 10
SSL_PORT_RANGE: PortRange = PortRange(1800, 2000)
DEFAULT_NODE_PRIVILEGES = "general"

DEFAULT_SDN_SOCKET_PORT = 1800

CATEGORIZED_LOG_LEVELS = ["ERROR", "WARNING"]
HAS_PREFIX = "with_logging_prefix"

FIBONACCI_GOLDEN_RATIO = (1 + 5 ** .5) / 2



import struct
from datetime import datetime
from typing import Optional, Dict

from dataclasses import dataclass

from bxcommon import constants
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.network.ip_endpoint import IpEndpoint
from bxcommon.utils.stats import message_utils
from bxutils.logging import LogLevel
from bxutils import logging

logger = logging.get_logger(__name__)

@dataclass
class BdnPerformanceStatsData:
    new_blocks_received_from_blockchain_node: int = 0
    new_blocks_received_from_bdn: int = 0
    new_blocks_seen: int = 0

    # block_messages vs. block_announcements might not be a distinction that
    # exists in all blockchains. For example, in Ethereum this is the
    # distinction between NewBlock and NewBlockHashes messages
    new_block_messages_from_blockchain_node: int = 0
    new_block_announcements_from_blockchain_node: int = 0

    new_tx_received_from_blockchain_node: int = 0
    new_tx_received_from_bdn: int = 0
    new_tx_received_from_blockchain_node_low_fee: int = 0  # not sent in msg
    new_tx_received_from_bdn_low_fee: int = 0  # not sent in msg

    tx_sent_to_node: int = 0
    duplicate_tx_from_node: int = 0


class BdnPerformanceStatsMessage(AbstractBloxrouteMessage):
    """
    Bloxroute message sent from gateway to relay that contains statistics on BDN performance.
    """
    MESSAGE_TYPE = BloxrouteMessageType.BDN_PERFORMANCE_STATS

    _interval_start_time: Optional[float] = None
    _interval_end_time: Optional[float] = None
    _memory_utilization_mb: Optional[int] = None
    _node_stats: Optional[Dict[IpEndpoint, BdnPerformanceStatsData]] = None

    def __init__(
        self,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        memory_utilization_mb: Optional[int] = None,
        node_stats: Optional[Dict[IpEndpoint, BdnPerformanceStatsData]] = None,
        buf: Optional[bytearray] = None
    ):
        self.node_stats_offset = (
            AbstractBloxrouteMessage.HEADER_LENGTH
            + (1 * constants.UL_SHORT_SIZE_IN_BYTES)
            + (2 * constants.DOUBLE_SIZE_IN_BYTES)
        )

        if buf is None:
            assert start_time is not None
            assert end_time is not None
            assert memory_utilization_mb is not None
            assert node_stats is not None
            buf = self._serialize(start_time, end_time, memory_utilization_mb, node_stats)

        self.buf = buf
        payload_length = len(buf) - AbstractBloxrouteMessage.HEADER_LENGTH
        super().__init__(self.MESSAGE_TYPE, payload_length, self.buf)

    def log_level(self) -> LogLevel:
        return LogLevel.DEBUG

    def interval_start_time(self) -> datetime:
        if self._interval_start_time is None:
            self._unpack()

        interval_start_time = self._interval_start_time
        assert interval_start_time is not None
        return datetime.fromtimestamp(interval_start_time)

    def interval_end_time(self) -> datetime:
        if self._interval_end_time is None:
            self._unpack()

        interval_end_time = self._interval_end_time
        assert interval_end_time is not None
        return datetime.fromtimestamp(interval_end_time)

    def memory_utilization(self) -> int:
        if self._memory_utilization_mb is None:
            self._unpack()

        memory_utilization_mb = self._memory_utilization_mb
        assert memory_utilization_mb is not None
        return memory_utilization_mb

    def node_stats(self) -> Dict[IpEndpoint, BdnPerformanceStatsData]:
        if self._node_stats is None:
            self._unpack()

        node_stats = self._node_stats
        assert node_stats is not None
        return node_stats

    def _serialize(
        self,
        start_time: datetime,
        end_time: datetime,
        memory_utilization_mb: int,
        node_stats: Dict[IpEndpoint, BdnPerformanceStatsData]
    ):
        stats_serialized_length = constants.UL_SHORT_SIZE_IN_BYTES + len(node_stats) * (
            constants.IP_ADDR_SIZE_IN_BYTES +
            constants.UL_SHORT_SIZE_IN_BYTES +
            (2 * constants.UL_SHORT_SIZE_IN_BYTES) +
            (7 * constants.UL_INT_SIZE_IN_BYTES)
        )
        msg_size = (
            self.node_stats_offset
            + stats_serialized_length
            + constants.CONTROL_FLAGS_LEN
        )
        buf = bytearray(msg_size)
        off = AbstractBloxrouteMessage.HEADER_LENGTH

        struct.pack_into("<d", buf, off, start_time.timestamp())
        off += constants.DOUBLE_SIZE_IN_BYTES

        struct.pack_into("<d", buf, off, end_time.timestamp())
        off += constants.DOUBLE_SIZE_IN_BYTES

        memory_utilization_mb = min(memory_utilization_mb, constants.UNSIGNED_SHORT_MAX_VALUE)
        struct.pack_into("<H", buf, off, memory_utilization_mb)
        off += constants.UL_SHORT_SIZE_IN_BYTES

        struct.pack_into("<H", buf, off, len(node_stats))
        off += constants.UL_SHORT_SIZE_IN_BYTES

        for endpoint, node_stat in node_stats.items():
            message_utils.pack_ip_port(buf, off, endpoint.ip_address, endpoint.port)
            off += constants.IP_ADDR_SIZE_IN_BYTES + constants.UL_SHORT_SIZE_IN_BYTES

            struct.pack_into("<H", buf, off, node_stat.new_blocks_received_from_blockchain_node)
            off += constants.UL_SHORT_SIZE_IN_BYTES

            struct.pack_into("<H", buf, off, node_stat.new_blocks_received_from_bdn)
            off += constants.UL_SHORT_SIZE_IN_BYTES

            struct.pack_into("<I", buf, off, node_stat.new_tx_received_from_blockchain_node)
            off += constants.UL_INT_SIZE_IN_BYTES

            struct.pack_into("<I", buf, off, node_stat.new_tx_received_from_bdn)
            off += constants.UL_INT_SIZE_IN_BYTES

            struct.pack_into("<I", buf, off, node_stat.new_blocks_seen)
            off += constants.UL_INT_SIZE_IN_BYTES

            struct.pack_into("<I", buf, off, node_stat.new_block_messages_from_blockchain_node)
            off += constants.UL_INT_SIZE_IN_BYTES

            struct.pack_into("<I", buf, off, node_stat.new_block_announcements_from_blockchain_node)
            off += constants.UL_INT_SIZE_IN_BYTES

            struct.pack_into("<I", buf, off, node_stat.tx_sent_to_node)
            off += constants.UL_INT_SIZE_IN_BYTES

            struct.pack_into("<I", buf, off, node_stat.duplicate_tx_from_node)
            off += constants.UL_INT_SIZE_IN_BYTES

        return buf

    def _unpack(self) -> None:
        buf = self.buf
        assert buf is not None
        node_stats = {}

        off = AbstractBloxrouteMessage.HEADER_LENGTH
        self._interval_start_time, = struct.unpack_from("<d", buf, off)
        off += constants.DOUBLE_SIZE_IN_BYTES
        self._interval_end_time, = struct.unpack_from("<d", buf, off)
        off += constants.DOUBLE_SIZE_IN_BYTES
        self._memory_utilization_mb, = struct.unpack_from("<H", buf, off)
        off += constants.UL_SHORT_SIZE_IN_BYTES

        num_peers, = struct.unpack_from("<H", buf, off)
        off += constants.UL_SHORT_SIZE_IN_BYTES
        for _ in range(num_peers):
            ip, port = message_utils.unpack_ip_port(self._memoryview[off:].tobytes())
            off += constants.IP_ADDR_SIZE_IN_BYTES + constants.UL_SHORT_SIZE_IN_BYTES
            single_node_stats = BdnPerformanceStatsData()
            single_node_stats.new_blocks_received_from_blockchain_node, = struct.unpack_from("<H", buf, off)
            off += constants.UL_SHORT_SIZE_IN_BYTES
            single_node_stats.new_blocks_received_from_bdn, = struct.unpack_from("<H", buf, off)
            off += constants.UL_SHORT_SIZE_IN_BYTES
            single_node_stats.new_tx_received_from_blockchain_node, = struct.unpack_from("<I", buf, off)
            off += constants.UL_INT_SIZE_IN_BYTES
            single_node_stats.new_tx_received_from_bdn, = struct.unpack_from("<I", buf, off)
            off += constants.UL_INT_SIZE_IN_BYTES
            single_node_stats.new_blocks_seen, = struct.unpack_from("<I", buf, off)
            off += constants.UL_INT_SIZE_IN_BYTES
            single_node_stats.new_block_messages_from_blockchain_node, = struct.unpack_from("<I", buf, off)
            off += constants.UL_INT_SIZE_IN_BYTES
            single_node_stats.new_block_announcements_from_blockchain_node, = struct.unpack_from("<I", buf, off)
            off += constants.UL_INT_SIZE_IN_BYTES
            single_node_stats.tx_sent_to_node, = struct.unpack_from("<I", buf, off)
            off += constants.UL_INT_SIZE_IN_BYTES
            single_node_stats.duplicate_tx_from_node, = struct.unpack_from("<I", buf, off)
            off += constants.UL_INT_SIZE_IN_BYTES
            node_stats[IpEndpoint(ip, port)] = single_node_stats
        self._node_stats = node_stats

    def __repr__(self) -> str:
        return (
            f"BdnPerformanceStatsMessage<"
            f"memory_utilization: {self.memory_utilization()}, "
            f"bdn_stats_per_node: {self.node_stats()}"
            f">"
        )



from typing import Callable, Optional, cast, List, Tuple

from bxcommon.utils.alarm_queue import AlarmQueue, AlarmId, Alarm


class MockAlarmQueue(AlarmQueue):
    def __init__(self) -> None:
        super().__init__()
        self.alarms = cast(
            List[Tuple[float, Callable]],
            self.alarms
        )

    # pylint: disable=arguments-differ
    def register_alarm(
        self, fire_delay: float, fn: Callable, *args, _alarm_name: Optional[str] = None, **kwargs
    ) -> AlarmId:
        self.alarms.append((fire_delay, fn))
        return AlarmId(fire_delay, 1, Alarm(fn, fire_delay))



from enum import Enum


class LogRecordType(Enum):
    Stats = "stats"
    BX = "bx"
    BlockInfo = "stats.blocks.events"
    BlockPropagationInfo = "stats.blocks.events.p"
    BdnPerformanceStats = "stats.bdn_performance"
    TransactionInfo = "stats.transactions.events"
    TransactionPropagationInfo = "stats.transactions.events.p"
    TransactionStats = "stats.transactions.summary"
    CloudApiPerformance = "stats.publicapi.performance"
    Throughput = "stats.throughput"
    Memory = "stats.memory"
    TransactionFeedStats = "stats.transaction.feed"
    EthOnBlockFeedStats = "stats.ethonblock.feed"
    NodeStatus = "stats.node.status"
    NodeInfo = "stats.node.info"
    NodeEvent = "stats.node.event"
    NetworkInfo = "stats.network.info"
    ConnectionState = "stats.connection_state"
    BlockCleanup = "bx.cleanup.block"
    TransactionCleanup = "bx.cleanup.transaction"
    BxMemory = "bx.memory"
    Recovery = "stats.recovery"
    TransactionHistogram = "bx.transaction.histogram"
    TransactionAudit = "stats.transaction.audit"
    TransactionAuditSummary = "stats.transaction.audit.summary"
    TransactionStatus = "transaction.status"
    TransactionTracing = "network_content.tx.tracing"
    BlockAuditSummary = "stats.block.audit.summary"
    AuditUpdates = "stats.audit.updates"
    CustomerInfo = "stats.customer.info"
    ExecutionTimerInfo = "stats.execution.timer.info"
    TaskDuration = "stats.task_duration"
    NetworkContent = "network_content.stats"
    NetworkContentTx = "network_content.tx.stats"
    NetworkContentBlock = "network_content.block.stats"
    RoutingService = "routing.service"
    ConnectionHealth = "stats.connection_health"
    PerformanceTroubleshooting = "stats.performance.responsiveness"
    MessageHandlingTroubleshooting = "stats.performance.message_handling"
    AlarmTroubleshooting = "stats.performance.alarm"
    NetworkTroubleshooting = "stats.performance.network"
    RoutingTableStats = "stats.routing"
    GarbageCollection = "stats.gc"
    ShortIdAllocation = "sid.allocation"
    Config = "config"
    QuotaNotification = "stats.quota.notification"
    QuotaFillStatus = "stats.quota.fill"
    TransactionFiltering = "transaction.filter"
    TxBlockchainStatusInfo = "stats.transaction.status"
    PrivateTransaction = "stats.private_transaction"
    PaidTransaction = "stats.transaction.paid_transaction"
    TxDetectionTimeLocation = "stats.transaction.tx_detection_time_location"
    TransactionMonitoring = "stats.transaction.monitoring"
    MEVBundleTransaction = "stats.transaction.mev_bundle"
    ProfitSharingPrivateTransaction = "stats.transaction.profit_sharing"
